def get_wmo_stations_info():
    """Return pandas.DataFrame with WMO stations information (WMO ids, coordinates, etc.)"""

    import pandas as pd

    wmo_ids_path = str(__file__).replace('wmo.py', 'wmo_ids_and_coords.csv')
    ids_coords = pd.read_csv(wmo_ids_path, dtype={3: 'object'}, sep=';', index_col=0)
    return ids_coords


def get_wmoid_or_coord(
        station_name, to_return, contains_station_name=True,
        station_name_is_wmo_id=False
):
    """
    Return WMO ids or coordinates of the WMO stations which contains 'station_name'
    in their names.

    Keyword arguments:
        station_name -- name of the station for which WMO id or coordinates will
    be returned. If 'cou' prefix added to 'station_name' and a country name appears
    after the prefix, the function will search for all stations in the specified
    country (e.g. 'couPOLAND' will return WMO ids or coordinates for all stations
    in Poland)
        to_return -- which parameter has to be returned ('wmo_id', 'lat', 'lon',
    'elv')
        contains_station_name -- if set to True, the function will search for
    stations which CONTAIN 'station_name' in their name. If set to False, the
    function will search for stations with NAMES EXACTLY as in 'station_name'
    (default True)
        station_name_is_wmo_id -- if set to True, the function will search for a
    station by its WMO id. In this case, in 'station_name' argument you have to
    pass WMO id
    """

    import pandas as pd

    station_name = str(station_name)
    wmo_ids_path = str(__file__).replace('wmo.py', 'wmo_ids_and_coords.csv')

    station_name = station_name.upper()
    ids_coords = pd.read_csv(wmo_ids_path, dtype={3: 'object'}, sep=';', index_col=0)
    if station_name.startswith('COU'):
        data = ids_coords[ids_coords['country'].str.contains(station_name.replace('COU', ''))]
    elif station_name_is_wmo_id:
        data = ids_coords[ids_coords['wmo_id'] == station_name]
    else:
        if contains_station_name:
            data = ids_coords[ids_coords['station'].str.contains(station_name)]
        else:
            data = ids_coords[ids_coords['station'] == station_name]

    if len(data.index) == 0:
        raise ValueError(
            """
            No data found for the given 'station_name'.
            """
        )
    elif station_name.startswith('COU'):
        indexes = [i for i in range(0, len(data.index))]
    elif len(data.index) > 1:
        stations_info = ''
        for i, station in enumerate(list(data.station)):
            stations_info += f'{i}: {list(data.country)[i]}, {list(data.wmo_id)[i]}, {station}\n            '

        indexes = input(
            f"""
            Data found for more than one station. Specify which station you are interested in (by selecting indexes. Use
            '*' if you are interested in every found station).            
            
            ---COUNTRY, WMO ID, STATION---
            {stations_info}
            """)

        if indexes == '*':
            all_indexes = []
            for i, station in enumerate(list(data.station)):
                all_indexes.append(i)
            indexes = all_indexes
        else:
            indexes = indexes.split(',')
        indexes = [int(i) for i in indexes]
    else:
        indexes = [0]

    dict_to_return = {}
    for index in indexes:
        dict_to_return[data.iloc[index, 1]] = data.iloc[index, :][to_return]

    return dict_to_return


def decode_downloaded_data(data_table):
    """
    Decode the data from .dat file from the WMO website. Return more
    computer-friendly data for further adaptation.
    """

    table = data_table.split('#')

    data = table[-1]
    row_to_del = data.split('\n')[0]
    data = data.replace(row_to_del, "")
    data = data.split('\n')

    decoded = []
    for row in data:
        dec_row = row.split(' ')
        while '' in dec_row:
            dec_row.remove('')
        decoded.append(dec_row)

    while [] in decoded:
        decoded.remove([])

    return decoded


def download_data(url):
    """Download and return the data from the given URL of the WMO website"""

    from bs4 import BeautifulSoup as bs
    import requests

    r = requests.get(url)
    soup = bs(r.content, features='html.parser')

    data_href = None
    for element in soup.find_all('a'):
        if '.dat' in element['href']:
            data_href = element['href']
            break

    if data_href is None:
        raise FileNotFoundError(
            f"""
            No data file was found for {url}.
            """
        )

    full_url_for_data = 'http://climexp.knmi.nl/' + data_href
    r = requests.get(full_url_for_data)
    soup = bs(r.content, features='html.parser')

    table = soup.get_text()
    downloaded_data = decode_downloaded_data(table)

    if not downloaded_data:
        raise FileNotFoundError(
            f"""
            No data file was found for {url}
            """
        )

    return downloaded_data


def concatenate_dfs(dfs):
    """Concatenate given dataframes into one consistent dataframe"""

    import pandas as pd

    min_year = 3000
    max_year = 0
    for df in dfs:
        if df[1][0] is not None:
            if int(df[1][0]) < min_year:
                min_year = int(df[1][0])
        if df[-1][0] is not None:
            if int(df[-1][0]) > max_year:
                max_year = int(df[-1][0])

    years = []
    months = []
    for year in range(min_year, max_year + 1):
        for month in range(1, 13):
            years.append(str(year))
            months.append(month)

    concatenated_df = pd.DataFrame({
        'year': years,
        'month': months,
    })

    for df in dfs:

        if df[1] == [None, None, None]:
            values_to_add = [None] * len(concatenated_df['month'])
            concatenated_df[df[0][2]] = values_to_add
        else:
            values_to_add = []

            index_to_start = None
            for i, year in enumerate(years):
                if year == df[1][0]:
                    index_to_start = i
                    break
                else:
                    values_to_add.append(None)

            if index_to_start is None:
                raise ValueError("'index_to_start' value is None.")

            for i, row in enumerate(df[1:]):
                if row[0] == years[index_to_start + i]:
                    if row[1] == months[index_to_start + i]:
                        if float(row[2]) > -90:
                            values_to_add.append(row[2])
                        else:
                            values_to_add.append(None)
                    else:
                        values_to_add.append(None)
                else:
                    values_to_add.append(None)

            while len(values_to_add) != len(years):
                values_to_add.append(None)

            concatenated_df[df[0][2]] = values_to_add

    for column in concatenated_df.columns[:2]:
        concatenated_df = concatenated_df.astype({column: int})
    for column in concatenated_df.columns[2:]:
        concatenated_df = concatenated_df.astype({column: float})

    return concatenated_df


def transpose_table(table):
    """Transpose the table for further adaptation"""

    new_table = []
    for row in table:
        for i, value in enumerate(row):
            if i != 0:
                new_row = [row[0], i, value]
                new_table.append(new_row)

    return new_table


def search_for_the_nearest_station(
        lat, lon, degrees_range=0.5
):
    """
    Return the nearest stations from the WMO database for the given coordinates.

    Keyword arguments:
        lat -- the latitude for which station will be searched
        lon -- the longitude for which station will be searched
        degrees_range -- acceptable range in degrees in all directions (default 0.5)
    """

    import pandas as pd

    wmo_ids_path = str(__file__).replace('wmo.py', 'wmo_ids_and_coords.csv')

    ids_coords = pd.read_csv(wmo_ids_path, dtype={3: 'object'}, sep=';', index_col=0)

    ids_coords = ids_coords[
        (ids_coords['lat'] < lat + degrees_range) &
        (ids_coords['lat'] > lat - degrees_range) &
        (ids_coords['lat'] != lat)
        ]

    ids_coords = ids_coords[
        (ids_coords['lon'] < lon + degrees_range) &
        (ids_coords['lon'] > lon - degrees_range) &
        (ids_coords['lon'] != lon)
        ]

    return ids_coords


def download_wmo_climatological_data(
        station_name, elements_to_scrape, nearby_stations=False,
        degrees_range_for_nearby_stations=0.5, return_coordinates=False
):
    """
    Download climatological data for specified station/stations from the WMO website.
    There are 5 elements which can be downloaded: average air temperature,
    sum of precipitation, absolute minimum air temperature, absolute maximum air
    temperature, sea level pressure. The interval for downloaded data is always
    monthly.

    Keyword arguments:
        station_name -- name of the station for which the data will to downloaded.
    If 'cou' prefix added to 'station_name' and a country name appears after the
    prefix, the function will search for all stations in the specified country
    (e.g. 'couPOLAND' will download data for Poland)
        elements_to_scrape -- which elements from the WMO website will be scraped.
    ('temp', 'preci', 'temp_min', 'temp_max', 'sl_press')
        nearby_stations -- if there is no data or a single element is missing, and
    if 'nearby_stations' argument is set to True, the function will search for the
    nearest stations and try to complete the lack (default False)
        degrees_range_for_nearby_stations -- acceptable range in degrees in all
    directions if nearby stations have to be searched (default 0.5)
        return_coordinates -- if set to True, the function will add columns with
    latitude, longitude and elevation for the specified station/stations (default False)
    """

    import pandas as pd

    if isinstance(elements_to_scrape, str):
        elements_to_scrape = [elements_to_scrape]

    elements_with_urls = {
        'temp': 'http://climexp.knmi.nl/gettempall.cgi?id=someone@somewhere&WMO={}',
        'preci': 'http://climexp.knmi.nl/getprcpall.cgi?id=someone@somewhere&WMO={}',
        'temp_min': 'http://climexp.knmi.nl/getminall.cgi?id=someone@somewhere&WMO={}',
        'temp_max': 'http://climexp.knmi.nl/getmaxall.cgi?id=someone@somewhere&WMO={}',
        'sl_press': 'http://climexp.knmi.nl/getslp.cgi?id=someone@somewhere&WMO={}'
    }

    wmo_ids = get_wmoid_or_coord(station_name, 'wmo_id')
    wmo_ids_list = [wmo_id for station, wmo_id in wmo_ids.items()]

    full_df = pd.DataFrame()
    print('Starting downloading data. It may take a while.')
    for station, wmo_id in wmo_ids.items():
        data = []
        for element in elements_to_scrape:
            url = elements_with_urls[element].format(wmo_id)

            try:
                downloaded_data = download_data(url)
            except FileNotFoundError:
                if 'cou' in station_name:
                    continue
                elif not nearby_stations:
                    print(
                        f"""
                        WARNING: No data for '{element}' in '{station}' (WMO ID: {wmo_id}).
                        If you want to search for data in the nearest stations, set 'nearby_stations'
                        argument to True.
                        """)
                    continue
                else:
                    lat = get_wmoid_or_coord(wmo_id, 'lat', station_name_is_wmo_id=True)[station]
                    lon = get_wmoid_or_coord(wmo_id, 'lon', station_name_is_wmo_id=True)[station]
                    nearest_stations = search_for_the_nearest_station(
                        lat, lon, degrees_range=degrees_range_for_nearby_stations
                    )
                    if nearest_stations.empty:
                        print(
                            f"""
                            WARNING: No data for '{element}' in '{station}' (WMO ID: {wmo_id}).
                            No nearby station found either.
                            """
                        )
                        continue

                    data_from_near_stations = {}
                    for near_wmo_id in nearest_stations['wmo_id']:
                        if near_wmo_id in wmo_ids_list:
                            continue

                        url = elements_with_urls[element].format(near_wmo_id)
                        try:
                            data_for_station = download_data(url)
                        except FileNotFoundError:
                            data_from_near_stations[near_wmo_id] = None
                        else:
                            data_from_near_stations[near_wmo_id] = data_for_station

                    hm_elements = 0
                    more_elements_in = None
                    for near_wmo_id, data_for_station in data_from_near_stations.items():
                        try:
                            if len(data_for_station) > hm_elements:
                                hm_elements = len(data_for_station)
                                more_elements_in = near_wmo_id
                        except TypeError:
                            continue

                    if more_elements_in is None:
                        downloaded_data = None
                    else:
                        downloaded_data = data_from_near_stations[more_elements_in]
                        print(
                            f"""
                            Warning: no '{element}' data was found for the chosen station ({station}), so the data was 
                            taken from the nearest station (WMO ID: {more_elements_in}). Latitude and longitude differences 
                            were below 0.5 degrees (default) or as in the 'degrees_range_for_nearby_stations' argument (if 
                            specified). If you do not want to download data from the nearest station, change 'nearby_stations' 
                            argument value to False. If you would like to change acceptable latitude and longitude differences, 
                            you can do it by passing float/int to 'degrees_range_for_nearby_stations' argument.
                            """)
            else:
                pass

            if downloaded_data is not None:
                downloaded_data = transpose_table(downloaded_data)
                downloaded_data.insert(0, ['year', 'month', element])
            else:
                downloaded_data = [[None, None, None]]
                downloaded_data.insert(0, ['year', 'month', element])

            data.append(downloaded_data)

        concatenated_df = concatenate_dfs(data)
        station_series = [station] * len(concatenated_df.index)
        concatenated_df.insert(0, 'station', station_series)

        if return_coordinates:
            lat = get_wmoid_or_coord(wmo_id, 'lat', station_name_is_wmo_id=True)[station]
            lon = get_wmoid_or_coord(wmo_id, 'lon', station_name_is_wmo_id=True)[station]
            elv = get_wmoid_or_coord(wmo_id, 'elv', station_name_is_wmo_id=True)[station]

            lat_series = [lat] * len(concatenated_df.index)
            lon_series = [lon] * len(concatenated_df.index)
            elv_series = [elv] * len(concatenated_df.index)

            concatenated_df['lat'] = lat_series
            concatenated_df['lon'] = lon_series
            concatenated_df['elv'] = elv_series

        full_df = full_df.append(concatenated_df)

        if 'cou' in station_name:
            columns_order = elements_to_scrape
            filtered_columns_order = ['station', 'year', 'month']

            for element in columns_order:
                if element in list(full_df.columns):
                    filtered_columns_order.append(element)

            if return_coordinates:
                filtered_columns_order += ['lat', 'lon', 'elv']
            full_df = full_df[filtered_columns_order]

    print('Data downloaded.')
    return full_df



import pytest
from cloupy.scraping import wmo
import mock
import builtins
from random import shuffle
import urllib.request
import urllib.error


def check_if_NOT_connected_to_the_internet(host='http://google.com'):
    try:
        urllib.request.urlopen(host)
        return False
    except urllib.error.URLError:
        return True


@pytest.mark.skipif(check_if_NOT_connected_to_the_internet(), reason='internet connection required')
class TestDataDownloading:

    @pytest.fixture
    def stations(self):
        return ['POZNAN', 'PERTH', 'HARARE', 'PARIS', 'BARCELONA', 'QUITO']

    @pytest.fixture
    def input_dict(self):
        return {
            'BARCELONA': '0',
            'PERTH': '1',
            'HARARE': '1',
            'QUITO': '1',
            'PARIS': '3'
        }

    @pytest.fixture
    def elements(self):
        return ['temp', 'preci', 'temp_max', 'temp_min', 'sl_press']

    def test_default_settings(
            self, stations, elements,
            input_dict
    ):
        for station in stations:
            shuffle(elements)
            if station != 'POZNAN':
                with mock.patch.object(builtins, 'input', lambda _: input_dict[station]):
                    data = wmo.download_wmo_climatological_data(station, elements)
                    assert not data.empty
                    assert TestDataDownloading.check_if_columns_in_proper_order(
                        list(data.columns[3:]), elements
                    )
                    for column in data.columns:
                        assert not data[column].isnull().all()
            else:
                data = wmo.download_wmo_climatological_data(station, elements)
                assert not data.empty
                assert TestDataDownloading.check_if_columns_in_proper_order(
                    list(data.columns[3:]), elements
                )
                for column in data.columns:
                    assert not data[column].isnull().all()

    def test_return_coordinates_arg(
            self, stations, elements,
            input_dict
    ):
        for station in stations:
            if station != 'POZNAN':
                with mock.patch.object(builtins, 'input', lambda _: input_dict[station]):
                    data = wmo.download_wmo_climatological_data(
                        station, elements, return_coordinates=True
                    )
                    assert list(data.columns[-1:-4:-1]) == ['elv', 'lon', 'lat']
                    for element in ['lat', 'lon', 'elv']:
                        assert not data[element].isnull().any()
            else:
                data = wmo.download_wmo_climatological_data(
                    station, elements, return_coordinates=True
                )
                assert list(data.columns[-1:-4:-1]) == ['elv', 'lon', 'lat']
                for element in ['lat', 'lon', 'elv']:
                    assert not data[element].isnull().any()

    def test_nearby_stations_arg(
            self, elements
    ):
        stations_and_input_val = {
            'PARIS': '4',
            'NEW YORK': '1',
            'MADRID': '1',
            'KOLO': '1'
        }

        for station, input_val in stations_and_input_val.items():
            with mock.patch.object(builtins, 'input', lambda _: input_val):
                if station == 'KOLO':
                    data = wmo.download_wmo_climatological_data(
                        station, elements, nearby_stations=True,
                        degrees_range_for_nearby_stations=1
                    )
                else:
                    data = wmo.download_wmo_climatological_data(station, elements, nearby_stations=True)
                assert list(data.columns[3:]) == elements
                assert TestDataDownloading.check_if_columns_in_proper_order(
                    list(data.columns[3:]), elements
                )

    def test_downloading_for_countries(
            self, elements
    ):
        countries = [
            'couICELAND', 'couCONGO', 'couVENEZUELA',
            'couPARAGUAY', 'couURUGUAY', 'couCUBA'
        ]

        for country in countries:
            shuffle(elements)
            data = wmo.download_wmo_climatological_data(country, elements)

            for column in data.columns:
                assert not data[column].isnull().all()
            assert list(data.columns)[3:] == elements

    @staticmethod
    def check_if_columns_in_proper_order(
            columns, elements_
    ):
        filtered_elements = []
        for element in elements_:
            if element in columns:
                filtered_elements.append(element)

        if columns == filtered_elements:
            return True
        else:
            return False



class WalterLieth:
    """
    Create a WalterLieth object in which data for drawing a Walter-Lieth diagram
    can be downloaded, modified, manually provided.

    Keyword arguments:
    station_name -- name of the station for which data will be drawn
    years_range -- years range for which data will be drawn (default None)
    dataframe -- data for the drawing (default None)
    lat -- latitude of the station (default None)
    lon -- longitude of the station (default None)
    elevation -- elevation of the station (default None)

    ---------------METHODS---------------
    draw()
    d_imgw_data()
    d_wmo_data()
    import_global_df()
    -------------------------------------

    ---------------DATA STRUCTURE---------------
    The supported data structure for WalterLieth.dataframe is 5 or 6 columns of
    pandas.DataFrame object, depending on the data interval. The supported data
    intervals are daily and monthly.

    If the data interval is monthly, 5 columns are required:
    1st column: months,
    2nd column: average air temperature,
    3rd column: sum of precipitation,
    4th column: absolute maximum air temperature,
    5th column: absolute minimum air temperature

    If the data interval is daily, 6 columns are required:
    1st column: years,
    2nd column: months,
    3rd column: average air temperature,
    4th column: sum of precipitation,
    5th column: absoulte maximum air temperature,
    6th column: absolute minimum air temperature

    Exemplary dataframe:
    import pandas as pd
    dataframe = pd.DataFrame({
                'months': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
                'temp': [-2, -1, 0, 7, 15, 18, 19, 20, 18, 14, 8, 3],
                'preci': [50, 25, 55, 60, 70, 80, 90, 80, 68, 50, 45, 49],
                'max_temp': [10, 15, 17, 18, 19, 20, 35, 34, 25, 20, 15, 10],
                'min_temp': [-36, -29, -20, -15, -5, -1, 1, 2, -1, -4, -18, -22]
                            })
    --------------------------------------------

    ---------------NOTE THAT---------------
    The data must be provided for every month. Otherwise, the Walter-Lieth diagram
    can not be drawn.
    ---------------------------------------
    """

    def __init__(
            self, station_name, years_range=None,
            dataframe=None, lat=False, lon=False,
            elevation=False
    ):
        self.dataframe = dataframe
        self.years_range = years_range
        self.station_name = station_name.upper()
        self.lat = lat
        self.lon = lon
        self.elevation = elevation

    def draw(
            self, figsize=(7.74, 7.74), language=None,
            freeze_rectangles=True, title_text=True, years_text=True,
            coordinates_box=True, yearly_means_box=True, extremes_box=True,
            legend_box=True
    ):
        """
        Specify which elements have to be drawn and draw Walter-Lieth diagram.

        Keyword arguments:
            figsize -- figure size (default (7.74, 7.74))
            language -- choose language (None, 'POL', 'ENG') (default None). If
        None, then choose default language, which is 'ENG'
            freeze_rectangles -- if bottom rectangles which show freeze periods
        have to be drawn (default True)
            title_text -- if station name has to be drawn (default True)
            years_text -- if years range has to be drawn (default True)
            coordinates_box -- if box with coordinates info has to be drawn
        (default True)
            yearly_means_box -- if box with yearly means has to be drawn (default
        True)
            extremes_box -- if box with extreme temperatures has to be drawn
        (default True)
            legend_box -- if legend has to be drawn (default True)
        """

        import matplotlib.pyplot as plt
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        from matplotlib import rcParams

        if self.dataframe is None or self.dataframe.empty:
            raise AttributeError(
                """
                No data in WalterLieth.dataframe. Input the data yourself or use the WalterLieth's methods: 
                WalterLieth.d_imgw_data, WalterLieth.d_wmo_data, WalterLieth.import_global_df
                """)

        ls_prop_cycle = rcParams['axes.prop_cycle'].by_key()['linestyle']
        color_prop_cycle = rcParams['axes.prop_cycle'].by_key()['color']
        if WalterLieth.check_cloudy_graphs_chosen_style(ls_prop_cycle, color_prop_cycle) == 'retro':
            temp_linecolor = 'k'
            temp_linestyle = '-.'

            preci_linecolor = 'k'
            preci_linestyle = '-'

            zero_linecolor = 'k'
            zero_linestyle = 'dotted'

            freeze_color = '#767676'
            likely_freeze_color = '#b6b6b6'

            dry_period_hatch = '..'
            dry_period_hatch_color = '#b6b6b6'

            wet_period_hatch = '||'
            wet_period_hatch_color = 'k'

        else:  # default or customized
            temp_linecolor = 'r'
            temp_linestyle = '-'

            preci_linecolor = '#1f77b4'
            preci_linestyle = '-'

            zero_linecolor = 'k'
            zero_linestyle = 'dotted'

            freeze_color = '#09d1d9'
            likely_freeze_color = '#c8f8fa'  # b3fcff

            dry_period_hatch = '..'
            dry_period_hatch_color = 'r'

            wet_period_hatch = '||'
            wet_period_hatch_color = '#1f77b4'

        if language == 'POL':
            asl = 'Wysokość: {} m n.p.m.'
            lat = 'Szerokość geograficzna: {}'
            lon = 'Długość geograficzna: {}'
            temp_label = 'Średnia roczna temperatura: {}°C'
            preci_label = 'Roczna suma opadów: {} mm'
            x_ticks_labels = ['S', 'L', 'M', 'K', 'M', 'C', 'L', 'S', 'W', 'P', 'L', 'G']
            temp_axis_ylabel = 'Średnia miesięczna temperatura powietrza [°C]'
            if freeze_rectangles:
                temp_axis_xlabel = 'Miesiące z prawdopodobieństwem wystąpienia przymrozku'
            else:
                temp_axis_xlabel = 'Miesiące'
            preci_axis_ylabel = 'Średnia miesięczna suma opadów [mm]'
            freeze_legend_label = 'przymrozek'
            likely_freeze_legend_label = 'możliwy przymr.'
            temp_legend_label = 'temp. powietrza'
            preci_legend_label = 'opad atmosf.'
            humid_period_legend_label = 'okres wilg.'
            wet_period_legend_label = 'okres mokry'
            dry_period_legend_label = 'okres suchy'
        else:
            asl = 'Elevation: {} m a.s.l.'
            lat = 'Latitude: {}'
            lon = 'Longitude: {}'
            temp_label = 'Average air temperature: {}°C'
            preci_label = 'Annual precipitation: {} mm'
            x_ticks_labels = ['J', 'F', 'M', 'A', 'M', 'J', 'J', 'A', 'S', 'O', 'N', 'D']
            temp_axis_ylabel = 'Average monthly air temperature [°C]'
            if freeze_rectangles:
                temp_axis_xlabel = 'Months and probability of freeze occurrence'
            else:
                temp_axis_xlabel = 'Months'
            preci_axis_ylabel = 'Average monthly precipitation sum [mm]'
            freeze_legend_label = 'freeze'
            likely_freeze_legend_label = 'likely freeze'
            temp_legend_label = 'air temp.'
            preci_legend_label = 'precipit.'
            humid_period_legend_label = 'humid period'
            wet_period_legend_label = 'wet period'
            dry_period_legend_label = 'dry period'

        title = self.station_name
        if self.years_range is None:
            years = False
        elif len(self.years_range) == 1:
            years = self.years_range[0]
        else:
            years = f'{str(min(self.years_range))}-{str(max(self.years_range))}'

        if self.lat:
            lat = lat.format(self.lat)

        if self.lon:
            lon = lon.format(self.lon)

        if self.elevation:
            asl = asl.format(self.elevation)

        #  data converting
        if len(self.dataframe.columns) == 5:
            mean_temperature = self.dataframe.groupby(self.dataframe.columns[0]).mean().iloc[:, 0]
            precipitation = self.dataframe.groupby(self.dataframe.columns[0]).mean().iloc[:, 1]
            abs_max_temp = self.dataframe.groupby(self.dataframe.columns[0]).max().iloc[:, 2]
            abs_min_temp = self.dataframe.groupby(self.dataframe.columns[0]).min().iloc[:, 3]
        elif len(self.dataframe.columns) == 6:
            mean_temperature = self.dataframe.groupby(self.dataframe.columns[1]).mean().iloc[:, 1]
            precipitation = self.dataframe.groupby([self.dataframe.columns[0], self.dataframe.columns[1]]).sum()
            precipitation = precipitation.groupby(self.dataframe.columns[1]).mean().iloc[:, 1]
            abs_max_temp = self.dataframe.groupby(self.dataframe.columns[1]).max().iloc[:, 3]
            abs_min_temp = self.dataframe.groupby(self.dataframe.columns[1]).min().iloc[:, 4]
        else:
            raise AttributeError(
                f"""
                Data in cloupy.WalterLieth.dataframe has {len(self.dataframe.columns)} columns which is invalid. Input 5 
                or 6 columns, depending on the data interval.
                """)

        if precipitation.min() < 0:
            raise ValueError(
                f"""
                A value below 0 mm occured in the precipitation series. Invalid value: {precipitation.min()}
                """)

        if mean_temperature.max() > 50:
            raise ValueError(
                f"""
                Monthly mean temperature exceeded 50°C! Invalid value: {mean_temperature.max()}
                """)

        annual_mean_temp = round(mean_temperature.mean(), 1)
        sum_preci = round(precipitation.sum(), 1)

        maxtwm_mintcm = WalterLieth.get_max_twm_and_min_tcm(mean_temperature, abs_max_temp, abs_min_temp)
        max_temp = round(maxtwm_mintcm[0], 1)
        min_temp = round(maxtwm_mintcm[1], 1)

        x_major_ticks = [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5]
        x_minor_ticks = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]

        temp_yaxis_ticks = WalterLieth.get_temp_yticks(mean_temperature)
        upper_axis_max_tick = WalterLieth.get_max_ytick_for_precipitation(precipitation)

        if freeze_rectangles:
            if min(temp_yaxis_ticks) < -45:
                y_value_for_closing_bottom_rectangles = min(temp_yaxis_ticks) + 2.7
            elif min(temp_yaxis_ticks) == -40:
                y_value_for_closing_bottom_rectangles = min(temp_yaxis_ticks) + 2.5
            elif min(temp_yaxis_ticks) == -30:
                y_value_for_closing_bottom_rectangles = min(temp_yaxis_ticks) + 2.3
            elif min(temp_yaxis_ticks) == -20:
                y_value_for_closing_bottom_rectangles = min(temp_yaxis_ticks) + 2
            elif min(temp_yaxis_ticks) == -10:
                y_value_for_closing_bottom_rectangles = min(temp_yaxis_ticks) + 1.65
            else:
                y_value_for_closing_bottom_rectangles = min(temp_yaxis_ticks) + 1.45
        else:
            y_value_for_closing_bottom_rectangles = None

        fig, temp_axis = plt.subplots(figsize=figsize)
        fig.patch.set_alpha(1)

        if title and title_text:
            fig.suptitle(title, x=0.14, size=16, ha='left', fontweight='bold')
        if self.elevation and coordinates_box:
            fig.text(0.14, 0.933, asl)
        if self.lat and coordinates_box:
            fig.text(0.14, 0.91, lat)
        if self.lon and coordinates_box:
            fig.text(0.14, 0.887, lon)
        if years and years_text:
            fig.text(0.88, 0.935, years, size=14, ha='right', fontweight='bold')
        if yearly_means_box:
            fig.text(0.88, 0.91, temp_label.format(annual_mean_temp), ha='right')
            fig.text(0.88, 0.887, preci_label.format(sum_preci), ha='right')
        if extremes_box:
            fig.text(0.08, 0.86, 'Max:', ha='center')
            fig.text(0.08, 0.86 - 0.023, f'{str(max_temp)}°C', ha='center')
            fig.text(0.08, 0.80, 'Min:', ha='center')
            fig.text(0.08, 0.80 - 0.023, f'{str(min_temp)}°C', ha='center')

        temp_axis.set_xticks(x_major_ticks)
        temp_axis.set_xticks(x_minor_ticks, minor=True)
        temp_axis.set_xticklabels(x_ticks_labels)
        temp_axis.set_xlim(0, max(x_minor_ticks))
        temp_axis.set_yticks(temp_yaxis_ticks)
        temp_axis.set_ylim(min(temp_yaxis_ticks), max(temp_yaxis_ticks))
        temp_axis.tick_params(axis='y', which='major', length=6)
        temp_axis.tick_params(axis='x', which='minor', length=0)
        temp_axis.tick_params(axis='x', which='major', length=0)

        if freeze_rectangles and y_value_for_closing_bottom_rectangles is not None:
            temp_axis.plot([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
                           [(k * 0) + y_value_for_closing_bottom_rectangles for k in range(0, 13)],
                           linewidth=1, color='k')

        for month in range(0, 13):
            temp_axis.plot([month, month],
                           [min(temp_yaxis_ticks), y_value_for_closing_bottom_rectangles],
                           color='k', linestyle='solid', linewidth=1)

        if min(mean_temperature) < 0:
            temp_axis.plot([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12], [k * 0 for k in range(0, 13)],
                           color=zero_linecolor, linewidth=1, linestyle=zero_linestyle)

        preci_axis = temp_axis.secondary_yaxis('right')
        preci_axis.set_yticks([tick for tick in temp_yaxis_ticks if tick >= 0])
        preci_axis.set_yticklabels([tick * 2 for tick in temp_yaxis_ticks if tick >= 0])
        preci_axis.tick_params(axis='both', which='major', length=6)

        divider = make_axes_locatable(temp_axis)
        upper_axis = divider.append_axes("top", size=1, pad=0)
        upper_axis.tick_params(axis='y', length=6)
        upper_axis.set_xticks(x_major_ticks)
        upper_axis.set_xticks(x_minor_ticks, minor=True)
        upper_axis.set_xticklabels(['', '', '', '', '', '', '', '', '', '', '', ''])
        upper_axis.set_xlim(0, max(x_minor_ticks))
        upper_axis.set_yticks([upper_axis_max_tick])
        upper_axis.set_yticks([], minor=True)
        upper_axis.set_ylim((100, upper_axis_max_tick))
        upper_axis.yaxis.tick_right()
        upper_axis.set_yticklabels([str(upper_axis_max_tick)])
        upper_axis.tick_params(axis='x', which='both', color=[0, 0, 0, 0])
        upper_axis.spines['top'].set_color('white')
        upper_axis.text(-0.01, upper_axis_max_tick + (upper_axis_max_tick * 0.015),
                        '_', ha='right', fontweight='bold', size=11).set_fontname('Times New Roman')
        temp_axis.set_ylabel(temp_axis_ylabel)
        temp_axis.set_xlabel(temp_axis_xlabel)
        preci_axis.set_ylabel(preci_axis_ylabel)

        x_for_plotting = [0, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5, 12]
        interpolated_margins = WalterLieth.interpolate_margins(mean_temperature, precipitation)
        mean_temperature = interpolated_margins[0]
        precipitation = interpolated_margins[1]

        temp_axis.plot(x_for_plotting, mean_temperature, color=temp_linecolor, linestyle=temp_linestyle,
                       label=temp_legend_label)
        temp_axis.plot(x_for_plotting, [p / 2 for p in precipitation], color=preci_linecolor, linestyle=preci_linestyle,
                       label=preci_legend_label)
        upper_axis.plot(x_for_plotting, precipitation, color=preci_linecolor, linestyle=preci_linestyle,
                        label=preci_legend_label)

        periods = WalterLieth.distinguish_between_wet_and_dry(mean_temperature, precipitation)
        wet_periods = periods[0]
        dry_periods = periods[1]

        for period in wet_periods:
            where = [False, False, False, False,
                     False, False, False, False,
                     False, False, False, False,
                     False, False]
            for index in range(period[0], period[1] + 1):
                where[index] = True
            temp_axis.fill_between(x_for_plotting, [p / 2 for p in precipitation],
                                   [temp if temp > 0 else 0 for temp in mean_temperature],
                                   where=where, interpolate=True, hatch=wet_period_hatch,
                                   color='none', edgecolor=wet_period_hatch_color, linewidth=0.0,
                                   label=humid_period_legend_label)
            if max(precipitation) > 100:
                upper_axis.fill_between(x_for_plotting, precipitation, 100,
                                        where=where, interpolate=True, hatch=None,
                                        color=wet_period_hatch_color, label=wet_period_legend_label)

        for period in dry_periods:
            where = [False, False, False, False,
                     False, False, False, False,
                     False, False, False, False,
                     False, False]
            for index in range(period[0], period[1] + 1):
                where[index] = True
            temp_axis.fill_between(x_for_plotting, [p / 2 for p in precipitation], mean_temperature,
                                   where=where, interpolate=True, hatch=dry_period_hatch,
                                   color='none', edgecolor=dry_period_hatch_color, linewidth=0.0,
                                   label=dry_period_legend_label)

        if freeze_rectangles:
            if_freeze = WalterLieth.determine_if_freeze(abs_min_temp)
            likely = if_freeze[0]
            freeze = if_freeze[1]

            temp_axis.fill_between([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
                                   [k * 0 + min(temp_yaxis_ticks) for k in range(0, 13)],
                                   y_value_for_closing_bottom_rectangles,
                                   color='white',
                                   )

            if freeze:
                where = [False, False, False, False,
                         False, False, False, False,
                         False, False, False, False,
                         False]
                for index in freeze:
                    where[index] = True
                    where[index + 1] = True
                temp_axis.fill_between([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
                                       [k * 0 + min(temp_yaxis_ticks) for k in range(0, 13)],
                                       y_value_for_closing_bottom_rectangles,
                                       where=where, color=freeze_color,
                                       label=freeze_legend_label)

            if likely:
                where = [False, False, False, False,
                         False, False, False, False,
                         False, False, False, False,
                         False]
                for index in likely:
                    where[index] = True
                    where[index + 1] = True
                temp_axis.fill_between([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
                                       [k * 0 + min(temp_yaxis_ticks) for k in range(0, 13)],
                                       y_value_for_closing_bottom_rectangles,
                                       where=where, color=likely_freeze_color,
                                       label=likely_freeze_legend_label)

        if legend_box:
            handles, labels = temp_axis.get_legend_handles_labels()
            if max(precipitation) > 100:
                handles.append(upper_axis.get_legend_handles_labels()[0][1])
                labels.append(upper_axis.get_legend_handles_labels()[1][1])

            doubled_labels = []
            for label in labels:
                if labels.count(label) > 1:
                    doubled_labels.append(label)

            for label in doubled_labels:
                while labels.count(label) > 1:
                    labels.remove(label)
                    handles.remove(handles[labels.index(label)])

            if len(labels) % 3 == 0:
                NCOL = 3
            elif len(labels) % 2 == 0:
                NCOL = 2
            elif len(labels) % 5 == 0:
                NCOL = 5
            else:
                NCOL = 4

            fig.legend(handles, labels, loc='lower center', ncol=NCOL, bbox_to_anchor=(0.52, 0))

    def d_imgw_data(
            self, years_range, interval='monthly',
            stations_kind='synop', filter_station=True, check_years=True,
            return_coordinates=True
    ):
        """
        Download data for a drawing from the IMGW database.

        Keyword arguments:
            years_range -- years range (eg. range(2010, 2021))
            interval -- the data interval ('monthly', 'daily', 'prompt') (default
        'monthly')
            stations_kind -- stations kind from the IMGW database ('synop', 'climat',
        'fall') (default 'synop')
            filter_station -- if the downloaded data must be filtered by
        WalterLieth.station_name (default True)
            check_years -- check if the years range in the downloaded data matches
        the years range from WalterLieth.years_range. If not, change
        WalterLieth.years_range depending on years range in the downloaded data
        (default True)
            return_coordinates -- search for coordinates for the chosen station
        name in WalterLieth.station_name (default True)
        """

        from cloupy.scraping import imgw as imgw_scraping

        if interval == 'monthly':
            data = imgw_scraping.download_imgw_climatological_data(
                years_range=years_range, interval=interval,
                stations_kind=stations_kind, file_format_index=0,
                keywords=['Rok', 'Nazwa stacji', 'Miesiąc',
                          'Średnia temperatura miesięczna [°C]',
                          'Miesieczna suma opadów [mm]',
                          'Absolutna temperatura maksymalna [°C]',
                          'Absolutna temperatura minimalna [°C]'],
                return_coordinates=return_coordinates
            )
            if filter_station:
                data = data[data['Nazwa stacji'] == self.station_name]

            if check_years:
                max_year = data.loc[:, 'Rok'].max()
                min_year = data.loc[:, 'Rok'].min()
                self.years_range = range(min_year, max_year + 1)

            data = data.drop(['Nazwa stacji', 'Rok'], axis=1)
            if data.empty:
                raise ValueError(
                    """
                    No data found for the specified parameters. Check if the input arguments in the 'WalterLieth.d_imgw_data' 
                    method are valid.
                    """)

        elif interval == 'daily':
            if not filter_station:
                raise ValueError(
                    """
                    'cloupy.WalterLieth.d_imgw_data' does not support given combination: interval='daily', 
                    filter_station=False. Use interval='monthly' instead or stick with the daily interval and change 
                    the 'filter_station' argument to True.
                    """)

            data = imgw_scraping.download_imgw_climatological_data(
                years_range=years_range, interval=interval,
                stations_kind=stations_kind, file_format_index=0,
                keywords=['Rok', 'Nazwa stacji', 'Miesiąc',
                          'Średnia temperatura dobowa [°C]',
                          'Suma dobowa opadów [mm]',  # w k_d jest 'ów'
                          'Suma dobowa opadu [mm]',  # w s_d jest 'u'. dziwne.
                          'Maksymalna temperatura dobowa [°C]',
                          'Minimalna temperatura dobowa [°C]'],
                return_coordinates=return_coordinates
            )

            data = data[data['Nazwa stacji'] == self.station_name]
            if check_years:
                max_year = data.loc[:, 'Rok'].max()
                min_year = data.loc[:, 'Rok'].min()
                self.years_range = range(min_year, max_year + 1)

            data = data.drop(['Nazwa stacji'], axis=1)
            if data.empty:
                raise ValueError(
                    """
                    No data found for the specified parameters. Check if the input arguments in the 'WalterLieth.d_imgw_data' 
                    method are valid.
                    """)
        else:
            raise ValueError(
                """
                    No data found for the specified parameters. Check if the input arguments in the 'WalterLieth.d_imgw_data' 
                    method are valid. Check the 'interval' argument.
                    """)

        if return_coordinates:
            self.lat = round(data['lat'].mean(), 1)
            self.lon = round(data['lon'].mean(), 1)
            self.elevation = round(data['elv'].mean())
            data = data.drop(['lat', 'lon', 'elv'], axis=1)

        self.dataframe = data

    def d_wmo_data(
            self, nearby_stations=True, return_coordinates=True,
            degrees_range_for_nearby_stations=0.5, check_years=True
    ):
        """
        Download data for a drawing from the WMO database.

        Keyword arguments:
            nearby_stations -- if there is no data or a single element is missing,
        and if 'nearby_stations' argument is set to True, the function will search
        for the nearest stations and try to complete the lack (default True)
            return_coordinates -- if set to True, the function will add columns
        with latitude, longitude and elevation for the specified station/stations
        (default False)
            degrees_range_for_nearby_stations -- acceptable range in degrees in all
        directions if nearby stations have to be searched (default 0.5)
            check_years -- check if the years range in the downloaded data matches
        the years range from WalterLieth.years_range. If not, change
        WalterLieth.years_range depending on the years range in the downloaded data
        (default True)
        """

        from cloupy.scraping import wmo as wmo_scraping

        data = wmo_scraping.download_wmo_climatological_data(
            self.station_name, ['temp', 'preci', 'temp_max', 'temp_min'], nearby_stations=nearby_stations,
            return_coordinates=return_coordinates, degrees_range_for_nearby_stations=degrees_range_for_nearby_stations
        )

        if check_years:
            min_year = data['year'].min()
            max_year = data['year'].max()
            self.years_range = range(min_year, max_year + 1)

        data = data.drop(['station', 'year'], axis=1)

        if return_coordinates:
            lat = data['lat'].mean()
            lon = data['lon'].mean()
            elv = data['elv'].mean()

            self.lat = float(round(lat, 2))
            self.lon = float(round(lon, 2))
            self.elevation = int(round(elv))

            data = data.drop(['lat', 'lon', 'elv'], axis=1)

        for element in ['temp', 'preci', 'temp_max', 'temp_min']:
            try:
                data[element]
            except KeyError:
                if element == 'preci' or element == 'temp':
                    raise FileNotFoundError(
                        f"""
                        No essential data for '{element}' was downloaded - please, try again. If this does not fix the 
                        problem, there's a high probability that the required data is missing on the WMO website.
                        """
                    )
                else:
                    data[element] = [None] * len(data.index)
            else:
                continue

        for element in ['temp', 'preci', 'temp_max', 'temp_min']:
            if data[element].isnull().all():
                if element == 'temp' or element == 'preci':
                    raise ValueError(
                        f"""
                        No required data for '{element}'. It seems like the data was downloaded, but actually the column 
                        contains only None values.
                        """
                    )
                elif element == 'temp_min':
                    print(
                        f"""
                        WARNING: no data found for '{element}'. It is not required for drawing, but the box with extreme
                        temperatures will show None value. The probability of freeze occurence neither can be calculated,
                        so the rectangles will be blank. You can tell drawing function not to draw above elements:
                        freeze_rectangles=False, extreme_box=False.
                        """
                    )
                else:
                    print(
                        f"""
                        WARNING: no data found for '{element}'. It is not required for drawing, but the box with extreme
                        temperatures will show None value. You can tell drawing function not to draw box with extreme 
                        temperatures: extreme_box=False.
                        """
                    )
        self.dataframe = data

    def import_global_df(
            self, columns_order, filter_station=True,
            check_years=True, station_in_column=False, years_in_column=False
    ):
        """
        Import data for WalterLieth.dataframe from the global dataframe.

        Keyword arguments:
            columns_order -- specify which columns from the global dataframe are
        to be taken (list of indexes, 'imgw_monthly', 'imgw_daily'). If the global
        dataframe comes from the IMGW database and its structure has not been
        modified in any way, then 'imgw_monthly' or 'imgw_daily' can be used -
        depending on the data interval
            filter_station -- if the imported data must be filtered by
        WalterLieth.station_name (default True)
            check_years -- check if the years range in the global DataFrame matches
        WalterLieth.years_range. If not, change WalterLieth.years_range depending
        on the years range in the global DataFrame (default True)
            station_in_column -- if you want to filter the imported data by
        WalterLieth.station_name, specify the column in which station names are
        located (with an index which is an integer) (default False, accepts integers)
            years_in_column -- if you want to check if WalterLieth.years_range matches
        the years range in the imported global dataframe, specify the index of
        the column in which years data is located (default False, accepts integers)

        ---------------NOTE THAT---------------
        If the data in the global dataframe comes from the IMGW database and its
        structure has not been modified in any way, you can use 'imgw_monthly',
        'imgw_daily' in the 'columns_order' argument. If this is your case, you do
        not have to set the 'station_in_column' and 'years_in_column' arguments -
        the method will do it for you.

        If the data in the global dataframe does not come from the IMGW database,
        you have to specify the columns in which the required data for the drawing
        is located (by passing a list of integers to 'columns_order'). Note that
        if you want to filter the data by station name (from WalterLieth.station_name)
        or check if the years range in the imported data matches WalterLieth.years_range,
        you also have to pass proper columns from the global dataframe to the
        'columns_order' argument. When you do that, you have to set the indexes of
        the columns in which the years or the station name are located ('station_in_column'
        and 'years_in_column' arguments). Remember that these indexes must be for
        a new dataframe which originates after the data filtering by 'columns_order'.

        For example:
        columns_order = [2, 3, 4, 5, 6, 7, 8], where:
        2nd column contains station names (the index in the new dataframe is 0),
        3rd column contains years (the new index is 1),
        4rd column contains months (the new index is 2),
        5rd column contains average air temperatures (the new index is 3),
        6rd column contains average sum of precipitation (the new index is 4),
        7rd column contains absolute maximum air temperature (the new index is 5),
        8rd column contains absolute minimum air temperature (the new index is 6)

        So, if you want to filter by station name and check if WalterLieth.years_range
        matches the data in the imported global dataframe, 'station_in_column' and
        'years_in_column' should be:
        station_in_column = int(0)
        years_in_column = int(1)

        After filtering, the method will drop these columns and pass the dataframe
        to WalterLieth.dataframe without the columns in which station names and
        years were. If the structure of the dataframe after droping the above columns
        matches the required data structure for the WalterLieth class, the graph
        should be drawn correctly (see the WalterLieth class docstring for the
        required data structure)
        ---------------------------------------
        """

        from cloupy import read_global_df

        if isinstance(columns_order, list) and not isinstance(columns_order[0], int):
            raise ValueError(
                """
                1. Use a list of ints to specify which columns from the global pandas.DataFrame are necessary for drawing.
                2. Remember that you can get accurate info on the required data structure in the cloupy.WalterLieth docs.
                3. Alternatively, if your data source is the IMGW database you can use the 'imgw_monthly' or 'imgw_daily' 
                strings as input.
                """)

        if columns_order != 'imgw_monthly' and columns_order != 'imgw_daily' and not isinstance(columns_order, list):
            raise ValueError(
                """
                Valid inputs for the 'columns_order' argument: 'imgw_monthly', 'imgw_daily', a list of ints
                If the data source is the IMGW database and the default columns order is preserved, you can input the
                'imgw_monthly' or 'imgw_daily' strings for the 'columns_order' argument (depending on the data interval).
                """)

        if not isinstance(station_in_column, int) or not isinstance(years_in_column, int):
            raise ValueError(
                """
                For the 'station_in_column' and 'years_in_column' arguments only integers are valid inputs.
                """
            )

        df = read_global_df()

        if 'imgw' in columns_order:
            if columns_order == 'imgw_monthly':
                df_for_object = df.iloc[:, [1, 2, 3, 12, 16, 4, 8]]
            else:
                df_for_object = df.iloc[:, [1, 2, 3, 9, 13, 5, 7]]

            if filter_station:
                df_for_object = df_for_object[df_for_object.iloc[:, 0] == self.station_name]
                if df_for_object.empty:
                    raise ValueError(
                        """
                        The dataframe has just been filtered and the output was empty. Check your inputs which may affect 
                        filtering.
                        
                        1. It is possible that the input for the 'station_name' argument in 'cloupy.WalterLieth' is invalid.
                        2. You may be looking for a station which is not available in the global pandas.DataFrame.
                        3. If you have changed default column order from the IMGW database, the function will return this 
                        exception.
                        4. Check if the global dataframe does not contain any unnecessary column. If you read a .csv file 
                        by pandas.read_csv() and set it as a global dataframe, the function may have added an unwanted
                        column to the dataframe. It is possible that passing int(0) for the 'index_col' argument 
                        (in the 'pandas.read_csv()' function) will handle the exception.
                        5. If you're willing not to filter data by the 'station_name' argument, you can set 'filter_station' 
                        argument to False. Note that it is not recommended to draw a Walter-Lieth diagram if the data does not 
                        come from a single station.
                        """
                    )

            if check_years:
                max_year = df_for_object.iloc[:, 1].max()
                min_year = df_for_object.iloc[:, 1].min()
                self.years_range = range(min_year, max_year + 1)

            df_for_object = df_for_object.drop(df_for_object.columns[0], axis=1)  # station names
            if columns_order == 'imgw_monthly':
                df_for_object = df_for_object.drop(df_for_object.columns[0], axis=1)  # years
        else:
            df_for_object = df.iloc[:, columns_order]

            if isinstance(station_in_column, int):

                if filter_station:
                    df_for_object = df_for_object[df_for_object.iloc[:, station_in_column] == self.station_name]
                    df_for_object.drop(df_for_object.columns[station_in_column], axis=1, inplace=True)
                else:
                    raise ValueError(
                        """
                        If you want to filter by station name in the specified column ('station_in_column'), please set
                        'filter_station' argument to True.
                        """
                    )

                if df_for_object.empty:
                    raise ValueError(
                        """
                        The dataframe after filtering is empty. Check if the WalterLieth.station_name is in global 
                        dataframe and if the 'station_in_column' argument is valid.
                        """
                    )

            if isinstance(years_in_column, int):

                if check_years:
                    if isinstance(station_in_column, int):
                        years_in_column -= 1

                    year_min = df_for_object.iloc[:, years_in_column].min()
                    year_max = df_for_object.iloc[:, years_in_column].max()

                    self.years_range = range(year_min, year_max + 1)
                    df_for_object.drop(df_for_object.columns[years_in_column], axis=1, inplace=True)
                else:
                    raise ValueError(
                        """
                        If you want to update the WalterLieth.years_range by the specified column ('years_in_column'), 
                        please set the 'check_years' argument to True.
                        """
                    )

        self.dataframe = df_for_object

    @staticmethod
    def check_cloudy_graphs_chosen_style(
            ls_prop_cycle, color_prop_cycle
    ):
        """Return which style has been set globally for cloupy"""

        if ls_prop_cycle.count('-') == 10:
            return 'default'
        elif ls_prop_cycle.count('-') == 1 and color_prop_cycle.count('k') == 6:
            return 'retro'
        else:
            return 'customized'

    @staticmethod
    def get_max_twm_and_min_tcm(
            mean_temperature, abs_max_temp, abs_min_temp
    ):
        """
        Return the absolute maximum of the warmest month and the absolute minimum
        of the coldest month
        """

        the_highest_temp = mean_temperature.max()
        the_lowest_temp = mean_temperature.min()

        the_warmest_month = None
        the_coldest_month = None
        for index, temp in enumerate(mean_temperature):
            if temp == the_highest_temp:
                the_warmest_month = index + 1
            if temp == the_lowest_temp:
                the_coldest_month = index + 1
        try:
            return abs_max_temp[the_warmest_month], abs_min_temp[the_coldest_month]
        except KeyError:
            raise ValueError(
                """
                Invalid data structure in the 'cloupy.WalterLieth.dataframe'. See the 'cloupy.WalterLieth' docs for more info.
                
                1. Check if you haven't mistaken columns order in the 'cloupy.WalterLieth.dataframe'. 
                2. It's also possible that you have tried to pass daily interval and there's a lack of required column/columns.
                Note that the daily interval data must contain 6 columns - for more info on the required data structure 
                see the 'cloupy.WalterLieth' docs. If you used 'cloupy.WalterLieth.import_global_df' method for data 
                which comes from the IMGW database, check the values passed to the 'columns_order' argument (if the values 
                'imgw_monthly' or 'imgw_daily' were used correctly, depending on the IMGW data interval).
                """)

    @staticmethod
    def get_temp_yticks(temperature):
        """Return a tick list for the WalterLieth graph"""

        available_temp_axis_ticks = [-50, -40, -30, -20, -10, 0, 10, 20, 30, 40, 50]
        min_temp = min(list(temperature))

        if available_temp_axis_ticks.count(min_temp) == 1:
            return available_temp_axis_ticks[available_temp_axis_ticks.index(min_temp) - 1:]

        temp_axis_ticks = []
        try:
            for avail_temp_tick in available_temp_axis_ticks:
                if avail_temp_tick < min_temp:
                    if available_temp_axis_ticks[available_temp_axis_ticks.index(avail_temp_tick) + 1] > min_temp:
                        temp_axis_ticks.append(avail_temp_tick)
                    continue
                else:
                    temp_axis_ticks.append(avail_temp_tick)
        except IndexError:
            raise AttributeError(
                """
                Invalid data structure in the 'cloupy.WalterLieth.dataframe'. Check the 'cloupy.WalterLieth' docs for more 
                info on the required data structure.
                
                If you have downloaded the data yourself from the IMGW database and set it as a global dataframe, check 
                if you have chosen the correct file format to download. The available file formats for the 'cloupy.WalterLieth' 
                from the IMGW database are: ['s_m_d', 's_d'].
                
                cloupy.d_imgw_data(..., file_format='s_m_d')
                cloupy.d_imgw_data(..., file_format='s_d')
                """)

        if temp_axis_ticks[0] > 0:
            temp_axis_ticks = [0, 10, 20, 30, 40, 50]

        return temp_axis_ticks

    @staticmethod
    def get_yticks_labels_for_precipitation(temp_yaxis_ticks):
        """Return a list of labels for the secondary axis in the WalterLieth graph"""

        preci_ticks = []
        for index, tick in enumerate(temp_yaxis_ticks):
            if index == 0 and tick != 0:
                preci_ticks.append('')
            else:
                preci_ticks.append(tick * 2)

        return preci_ticks

    @staticmethod
    def get_max_ytick_for_precipitation(precipitation):
        """Return the highest tick for precipitation for the upper axis in the WL graph"""

        available_ticks = [200, 300, 400, 500, 600,
                           700, 800, 900, 1000, 1100,
                           1200, 1300, 1400, 1500, 1600,
                           1700, 1800, 1900, 2000]
        max_tick = max(list(precipitation))

        for tick in available_ticks:
            if tick >= max_tick:
                return tick
        raise ValueError(f"Precipitation is too high! {max_tick}")

    @staticmethod
    def determine_if_freeze(abs_min_temp):
        """Return a list of the months in which freeze is likely and more likely"""

        likely = []
        freeze = []

        for index, temp in enumerate(abs_min_temp):
            if temp < -5:
                freeze.append(index)
            elif temp < 0:
                likely.append(index)
            else:
                continue

        if len(likely) == 0:
            likely = False
        if len(freeze) == 0:
            freeze = False

        return likely, freeze

    @staticmethod
    def interpolate_margins(
            mean_temperature, precipitation
    ):
        """Interpolate the given data to the edges of the WalterLieth graph"""

        temperature = list(mean_temperature)
        precipitation = list(precipitation)

        if temperature[1] - temperature[0] > 0:
            temperature_trend = 'decrease'
        elif temperature[1] - temperature[0] == 0:
            temperature_trend = 'stable'
        else:
            temperature_trend = 'increase'

        if precipitation[1] - precipitation[0] > 0:
            precipitation_trend = 'decrease'
        elif precipitation[1] - precipitation[0] == 0:
            precipitation_trend = 'stable'
        else:
            precipitation_trend = 'increase'

        if temperature_trend == 'decrease':
            if temperature[0] < 0:
                temperature_0 = temperature[0] - ((temperature[0] * 0.05) * -1)
            else:
                temperature_0 = temperature[0] - (temperature[0] * 0.05)
        elif temperature_trend == 'increase':
            if temperature[0] < 0:
                temperature_0 = temperature[0] + ((temperature[0] * 0.05) * -1)
            else:
                temperature_0 = temperature[0] + (temperature[0] * 0.05)
        else:
            temperature_0 = temperature[0]

        if precipitation_trend == 'decrease':
            precipitation_0 = precipitation[0] - (precipitation[0] * 0.05)
        elif precipitation_trend == 'increase':
            precipitation_0 = precipitation[0] + (precipitation[0] * 0.05)
        else:
            precipitation_0 = precipitation[0]

        new_temperature = []
        for index, temp in enumerate(temperature):
            if index == 0:
                new_temperature.append(temperature_0)
                new_temperature.append(temp)
            else:
                new_temperature.append(temp)
        new_temperature.append(temperature_0)

        new_precipitation = []
        for index, preci in enumerate(precipitation):
            if index == 0:
                new_precipitation.append(precipitation_0)
                new_precipitation.append(preci)
            else:
                new_precipitation.append(preci)
        new_precipitation.append(precipitation_0)

        return new_temperature, new_precipitation

    @staticmethod
    def distinguish_between_wet_and_dry(
            mean_temperature, precipitation
    ):
        """Distinguish wet and dry periods and return appropriate lists"""

        corrected_precipitation = []
        for preci in precipitation:
            if preci < 100:
                corrected_precipitation.append(preci / 2)
            else:
                corrected_precipitation.append(preci)

        dry_months_indexes = []
        wet_months_indexes = []

        for index, preci in enumerate(corrected_precipitation):
            if preci > mean_temperature[index]:
                wet_months_indexes.append(index)
            else:
                dry_months_indexes.append(index)

        wet_periods = []
        period = []
        for i, index in enumerate(wet_months_indexes):
            if i == 0:
                period.append(index)
            elif index - wet_months_indexes[i - 1] != 1:
                period.append(wet_months_indexes[i - 1])
                wet_periods.append((period[0], period[1]))
                period.clear()

                period.append(index)
            else:
                continue
        if len(period) == 1:
            wet_period_starts_with = period[0]
            wet_period_ends_with = wet_months_indexes[-1]
            wet_periods.append((wet_period_starts_with, wet_period_ends_with))

        dry_periods = []
        period = []
        for i, index in enumerate(dry_months_indexes):
            if i == 0:
                period.append(index)
            elif index - dry_months_indexes[i - 1] != 1:
                period.append(dry_months_indexes[i - 1])
                dry_periods.append((period[0], period[1]))
                period.clear()

                period.append(index)
            else:
                continue
        try:
            if len(period) == 1:
                dry_period_starts_with = period[0]
                dry_period_ends_with = dry_months_indexes[-1]
                dry_periods.append((dry_period_starts_with, dry_period_ends_with))
        except KeyError:
            pass

        return wet_periods, dry_periods



"""
Functions for IMGW database analysis.

get_meteorological_data(interval, stations_kind, years_range, file_format_index=0,
                        file_format=None,   specific_columns=None, keywords=None,
                        merge_splitted_stations=True, optimize_memory_usage=False)

get_file_formats(interval, stations_kind, file_format_index)
get_column_names(file_format)
look_for_keywords_in_columns(keywords, file_format=None)

get_urls(interval, stations_kind, years_range)
download_data(urls)
concatenate_data(downloaded_files_names, file_formats, specific_columns,
                keywords, optimize_memory_usage, years_range)
"""


def get_file_formats(
        interval, stations_kind, file_format_index
):
    """
    Return the available file formats for the given 'interval' and 'stations_kind'
    (different file formats contain different data).

    Keyword arguments:
        interval -- data interval from the IMGW database ('monthly', 'daily', 'prompt')
        stations_kind -- stations' kind ('synop', 'climat', 'fall')
        file_format_index -- which element from the list of available formats will
    be returned. Usually two file formats are available, so the length of the list
    is 2 ('all', 0, 1)
    """

    if interval == 'monthly':
        if stations_kind == 'synop':
            available_files_formats = ['s_m_d', 's_m_t']
        elif stations_kind == 'fall':
            available_files_formats = ['o_m']
        elif stations_kind == 'climat':
            available_files_formats = ['k_m_d', 'k_m_t']
        else:
            raise ValueError(
                "Invalid 'stations_kind' input. Available inputs: 'synop', 'fall', 'climat'.")

    elif interval == 'daily':
        if stations_kind == 'synop':
            available_files_formats = ['s_d', 's_d_t']
        elif stations_kind == 'fall':
            available_files_formats = ['o_d']
        elif stations_kind == 'climat':
            available_files_formats = ['k_d', 'k_d_t']
        else:
            raise ValueError(
                "Invalid 'stations_kind' input. Available inputs: 'synop', 'fall', 'climat'.")

    elif interval == 'prompt':
        if stations_kind == 'synop':
            available_files_formats = ['s_t']
        elif stations_kind == 'fall':
            raise NotADirectoryError("There's no '/dane_meteorologiczne/terminowe/opad/' directory in IMGW database.")
        elif stations_kind == 'climat':
            available_files_formats = ['k_t']
        else:
            raise ValueError(
                "Invalid 'stations_kind' input. Available inputs: 'synop', 'fall', 'climat'.")

    else:
        raise ValueError("Invalid 'interval' input. Available inputs: 'monthly', 'daily', 'prompt'.")

    if file_format_index == 'all':
        chosen_file_format = available_files_formats
    elif file_format_index == 0 or file_format_index == 1:
        try:
            chosen_file_format = [available_files_formats[file_format_index]]
        except IndexError:
            if len(available_files_formats) == 1:
                chosen_file_format = [available_files_formats[0]]
            else:
                raise Exception("Something's wrong with the file format.")
    else:
        raise ValueError(
            "{} file formats for the given 'interval' and 'stations_kind' available. Use index of the file format or 'all' "
            "for 'file_format_index' argument (if 'all', both file formats will be taken): {}".format(
                len(available_files_formats), available_files_formats
            )
        )
    return chosen_file_format


def get_column_names(file_format):
    """
    Return the column names for the given file format.

    Keyword arguments:
        file_format -- IMGW database file format (e.g. 's_m_t'). Available file
    formats: k_m_d, k_m_t, o_m, s_m_d, s_m_t, k_d, k_d_t, o_d, s_d, s_d_t, k_t,
    s_t
    """

    if file_format == 'k_m_d':
        return ['Kod stacji', 'Nazwa stacji', 'Rok', 'Miesiąc', 'Absolutna temperatura maksymalna [°C]',
                'Status pomiaru TMAX', 'Średnia temperatura maksymalna [°C]', 'Status pomiaru TMXS',
                'Absolutna temperatura minimalna [°C]', 'Status pomiaru TMIN', 'Średnia temperatura minimalna [°C]',
                'Status pomiaru TMNS', 'Średnia temperatura miesięczna [°C]', 'Status pomiaru STM',
                'Minimalna temperatura przy gruncie [°C]', 'Status pomiaru TMNG', 'Miesieczna suma opadów [mm]',
                'Status pomiaru SUMM', 'Maksymalna dobowa suma opadów [mm]', 'Status pomiaru OPMX',
                'Pierwszy dzień wystapienia opadu maksymalnego', 'Ostatni dzień wystąpienia opadu maksymalnego',
                'Maksymalna wysokość pokrywy śnieżnej [cm]', 'Status pomiaru PKSN', 'Liczba dni z pokrywą śnieżną',
                'Liczba dni z opadem deszczu', 'Liczba dni z opadem śniegu']

    elif file_format == 'k_m_t':
        return ['Kod stacji', 'Nazwa stacji', 'Rok', 'Miesiąc', 'Średnia miesięczna temperatura [°C]',
                'Status pomiaru TEMP', 'Średnia miesięczna wilgotność względna [%]', 'Status pomiaru WLGS',
                'Średnia miesięczna prędkość wiatru [m/s]', 'Status pomiaru FWS',
                'Średnie miesięczne zachmurzenie ogólne [oktanty]', 'Status pomiaru NOS']

    elif file_format == 'o_m':
        return ['Kod stacji', 'Nazwa stacji', 'Rok', 'Miesiąc', 'Miesięczna suma opadów [mm]', 'Status pomiaru SUMM',
                'Liczba dni z opadem śniegu', 'Status pomiaru LDS', 'Opad maksymalny [mm]', 'Status pomiaru MAXO',
                'Dzień pierwszy wystąpienia opadu maksymalnego', 'Dzień ostatni wystąpienia opadu maksymalnego',
                'Liczba dni z pokrywą śnieżną', 'Status pomiaru LDPS']

    elif file_format == 's_m_d':
        return ['Kod stacji', 'Nazwa stacji', 'Rok', 'Miesiąc', 'Absolutna temperatura maksymalna [°C]',
                'Status pomiaru TMAX', 'Średnia temperatura maksymalna [°C]', 'Status pomiaru TMXS',
                'Absolutna temperatura minimalna [°C]', 'Status pomiaru TMIN', 'Średnia temperatura minimalna [°C]',
                'Status pomiaru TMNS', 'Średnia temperatura miesięczna [°C]', 'Status pomiaru STM',
                'Minimalna temperatura przy gruncie [°C]', 'Status pomiaru TMNG', 'Miesieczna suma opadów [mm]',
                'Status pomiaru SUMM', 'Maksymalna dobowa suma opadów [mm]', 'Status pomiaru OPMX',
                'Pierwszy dzień wystapienia opadu maksymalnego', 'Ostatni dzień wystąpienia opadu maksymalnego',
                'Miesięczna suma usłonecznienia [godziny]', 'Status pomiaru SUUS',
                'Maksymalna wysokość pokrywy śnieżnej [cm]', 'Status pomiaru PKSN', 'Liczba dni z pokrywą śnieżną',
                'Status pomiaru PSDN', 'Liczba dni z opadem deszczu', 'Status pomiaru DESD',
                'Liczba dni z opadem śniegu',
                'Status pomiaru SNID', 'Liczba dni z opadem deszczu ze śniegiem', 'Status pomiaru DSND',
                'Liczba dni z gradem', 'Status pomiaru GRDD', 'Liczba dni z mgłą', 'Status pomiaru MGLD',
                'Liczba dni z zamgleniem', 'Status pomiaru ZAMD', 'Liczba dni z sadzią', 'Status pomiaru SADD',
                'Liczba dni z gołoledzią', 'Status pomiaru GOLD', 'Liczba dni z zamiecią śnieżną niską',
                'Status pomiaru ZAND', 'Liczba dni z zamiecią śnieżną wysoką', 'Status pomiaru ZAWD',
                'Liczba dni ze zmętnieniem', 'Status pomiaru ZMED', 'Liczba dni z wiatrem >= 10m/s',
                'Status pomiaru W10D', 'Liczba dni z wiatrem >15m/s', 'Status pomiaru W15D', 'Liczba dni z burzą',
                'Status pomiaru BURD', 'Liczba dni z rosą', 'Status pomiaru ROSD', 'Liczba dni ze szronem',
                'Status pomiaru SZRD']

    elif file_format == 's_m_t':
        return ['Kod stacji', 'Nazwa stacji', 'Rok', 'Miesiąc', 'Średnie miesięczne zachmurzenie ogólne [oktanty]',
                'Status pomiaru NOS', 'Średnia miesięczna prędkość wiatru [m/s]', 'Status pomiaru FWS',
                'Średnia miesięczna temperatura [°C]', 'Status pomiaru TEMP',
                'Średnie miesięczne ciśnienie pary wodnej [hPa]', 'Status pomiaru CPW',
                'Średnia miesięczna wilgotność względna [%]', 'Status pomiaru WLGS',
                'Średnie miesięczne ciśnienie na poziomie stacji [hPa]', 'Status pomiaru PPPS',
                'Średnie miesięczne ciśnienie na pozimie morza [hPa]', 'Status pomiaru PPPM', 'Suma opadu dzień [mm]',
                'Status pomiaru WODZ', 'Suma opadu noc [mm]', 'Status pomiaru WONO']

    elif file_format == 'k_d':
        return ['Kod stacji', 'Nazwa stacji', 'Rok', 'Miesiąc', 'Dzień', 'Maksymalna temperatura dobowa [°C]',
                'Status pomiaru TMAX', 'Minimalna temperatura dobowa [°C]', 'Status pomiaru TMIN',
                'Średnia temperatura dobowa [°C]', 'Status pomiaru STD', 'Temperatura minimalna przy gruncie [°C]',
                'Status pomiaru TMNG', 'Suma dobowa opadów [mm]', 'Status pomiaru SMDB', 'Rodzaj opadu [S/W/ ]',
                'Wysokość pokrywy śnieżnej [cm]', 'Status pomiaru PKSN']

    elif file_format == 'k_d_t':
        return ['Kod stacji', 'Nazwa stacji', 'Rok', 'Miesiąc', 'Dzień', 'Średnia dobowa temperatura [°C]',
                'Status pomiaru TEMP', 'Średnia dobowa wilgotność względna [%]', 'Status pomiaru WLGS',
                'Średnia dobowa prędkość wiatru [m/s]', 'Status pomiaru FWS',
                'Średnie dobowe zachmurzenie ogólne [oktanty]', 'Status pomiaru NOS']

    elif file_format == 'o_d':
        return ['Kod stacji', 'Nazwa stacji', 'Rok', 'Miesiąc', 'Dzień', 'Suma dobowa opadów [mm]',
                'Status pomiaru SMDB', 'Rodzaj opadu [S/W/ ]', 'Wysokość pokrywy śnieżnej [cm]', 'Status pomiaru PKSN',
                'Wysokość świeżo spadłego śniegu [cm]', 'Status pomiaru HSS', 'Gatunek śniegu [kod]',
                'Status pomiaru GATS', 'Rodzaj pokrywy śnieżnej [kod]', 'Status pomiaru RPSN']

    elif file_format == 's_d':
        return ['Kod stacji', 'Nazwa stacji', 'Rok', 'Miesiąc', 'Dzień', 'Maksymalna temperatura dobowa [°C]',
                'Status pomiaru TMAX', 'Minimalna temperatura dobowa [°C]', 'Status pomiaru TMIN',
                'Średnia temperatura dobowa [°C]', 'Status pomiaru STD', 'Temperatura minimalna przy gruncie [°C]',
                'Status pomiaru TMNG', 'Suma dobowa opadu [mm]', 'Status pomiaru SMDB', 'Rodzaj opadu [S/W/ ]',
                'Wysokość pokrywy śnieżnej [cm]', 'Status pomiaru PKSN', 'Równoważnik wodny śniegu [mm/cm]',
                'Status pomiaru RWSN', 'Usłonecznienie [godziny]', 'Status pomiaru USL',
                'Czas trwania opadu deszczu [godziny]', 'Status pomiaru DESZ', 'Czas trwania opadu śniegu [godziny]',
                'Status pomiaru SNEG', 'Czas trwania opadu deszczu ze śniegiem [godziny]', 'Status pomiaru DISN',
                'Czas trwania gradu [godziny]', 'Status pomiaru GRAD', 'Czas trwania mgły [godziny]',
                'Status pomiaru MGLA', 'Czas trwania zamglenia  [godziny]', 'Status pomiaru ZMGL',
                'Czas trwania sadzi [godziny]', 'Status pomiaru SADZ', 'Czas trwania gołoledzi [godziny]',
                'Status pomiaru GOLO', 'Czas trwania zamieci śnieżnej niskiej [godziny]', 'Status pomiaru ZMNI',
                'Czas trwania zamieci śnieżnej wysokiej [godziny]', 'Status pomiaru ZMWS',
                'Czas trwania zmętnienia [godziny]', 'Status pomiaru ZMET', 'Czas trwania wiatru >=10m/s [godziny]',
                'Status pomiaru FF10', 'Czas trwania wiatru >15m/s [godziny]', 'Status pomiaru FF15',
                'Czas trwania burzy  [godziny]', 'Status pomiaru BRZA', 'Czas trwania rosy  [godziny]',
                'Status pomiaru ROSA', 'Czas trwania szronu [godziny]', 'Status pomiaru SZRO',
                'Wystąpienie pokrywy śnieżnej [0/1]', 'Status pomiaru DZPS', 'Wystąpienie błyskawicy [0/1]',
                'Status pomiaru DZBL', 'Stan gruntu [Z/R]', 'Izoterma dolna [cm]', 'Status pomiaru IZD',
                'Izoterma górna [cm]', 'Status pomiaru IZG', 'Aktynometria  [J/cm2]', 'Status pomiaru AKTN']

    elif file_format == 's_d_t':
        return ['Kod stacji', 'Nazwa stacji', 'Rok', 'Miesiąc', 'Dzień', 'Średnie dobowe zachmurzenie ogólne [oktanty]',
                'Status pomiaru NOS', 'Średnia dobowa prędkość wiatru [m/s]', 'Status pomiaru FWS',
                'Średnia dobowa temperatura [°C]', 'Status pomiaru TEMP', 'Średnia dobowe ciśnienie pary wodnej [hPa]',
                'Status pomiaru CPW', 'Średnia dobowa wilgotność względna [%]', 'Status pomiaru WLGS',
                'Średnia dobowe ciśnienie na poziomie stacji [hPa]', 'Status pomiaru PPPS',
                'Średnie dobowe ciśnienie na pozimie morza [hPa]', 'Status pomiaru PPPM', 'Suma opadu dzień [mm]',
                'Status pomiaru WODZ', 'Suma opadu noc [mm]', 'Status pomiaru WONO']

    elif file_format == 'k_t':
        return ['Kod stacji', 'Nazwa stacji', 'Rok', 'Miesiąc', 'Dzień', 'Godzina', 'Temperatura powietrza [°C]',
                'Status pomiaru TEMP', 'Temperatura termometru zwilżonego [°C]', 'Status pomiaru TTZW',
                'Wskaźnik lodu [L/W]', 'Wskaźnik wentylacji [W/N]', 'Wilgotność względna [%]', 'Status pomiaru WLGW',
                'Kod kierunku wiatru [kod]', 'Status pomiaru DKDK', 'Prędkość wiatru [m/s]', 'Status pomiaru FWR',
                'Zachmurzenie ogólne [0-10 do dn.31.12.1988/oktanty od dn.01.01.1989]', 'Status pomiaru ZOGK',
                'Widzialność [kod]', 'Status pomiaru WID']

    elif file_format == 's_t':
        return ['Kod stacji', 'Nazwa stacji', 'Rok', 'Miesiąc', 'Dzień', 'Godzina',
                'Wysokość podstawy chmur CL CM szyfrowana [kod]', 'Status pomiaru HPOD',
                'Wysokość podstawy niższej [m]',
                'Status pomiaru HPON', 'Wysokość podstawy wyższej [m]', 'Status pomiaru HPOW',
                'Wysokość podstawy tekstowy [opis]', 'Pomiar przyrzadem 1 (niższa) [P]',
                'Pomiar przyrzadem 2 (wyższa) [P]',
                'Widzialność [kod]', 'Status pomiaru WID', 'Widzialność operatora [m]', 'Status pomiaru WIDO',
                'Widzialność automat [m]', 'Status pomiaru WIDA', 'Zachmurzenie ogólne [oktanty]', 'Status pomiaru NOG',
                'Kierunek wiatru  [°]', 'Status pomiaru KRWR', 'Prędkość wiatru [m/s]', 'Status pomiaru FWR',
                'Poryw wiatru [m/s]', 'Status pomiaru PORW', 'Temperatura powietrza [°C]', 'Status pomiaru TEMP',
                'Temperatura termometru zwilżonego [°C]', 'Status pomiaru TTZW', 'Wskaźnik wentylacji [W/N]',
                'Wskaźnik lodu [L/W]', 'Ciśnienie pary wodnej [hPa]', 'Status pomiaru CPW', 'Wilgotność względna [%]',
                'Status pomiaru WLGW', 'Temperatura punktu rosy [°C]', 'Status pomiaru TPTR',
                'Ciśnienie na pozimie stacji [hPa]', 'Status pomiaru PPPS', 'Ciśnienie na poziomie morza [hPa]',
                'Status pomiaru PPPM', 'Charakterystyka tendencji [kod]', 'Wartość tendencji [wartość]',
                'Status pomiaru APP',
                'Opad za 6 godzin [mm]', 'Status pomiaru WO6G', 'Rodzaj opadu za 6 godzin [kod]', 'Status pomiaru ROPT',
                'Pogoda bieżąca [kod]', 'Pogoda ubiegła [kod]', 'Zachmurzenie niskie [oktanty]', 'Status pomiaru CLCM',
                'Chmury CL [kod]', 'Status pomiaru CHCL', 'Chmury CL tekstem', 'Chmury CM [kod]', 'Status pomiaru CHCM',
                'Chmury CM tekstem', 'Chmury CH [kod]', 'Status pomiaru CHCH', 'Chmury CH tekstem', 'Stan gruntu [kod]',
                'Status pomiaru SGRN', 'Niedosyt wilgotności [hPa]', 'Status pomiaru DEFI', 'Usłonecznienie',
                'Status pomiaru USLN', 'Wystąpienie rosy [0/1]', 'Status pomiaru ROSW',
                'Poryw maksymalny za okres WW [m/s]', 'Status pomiaru PORK', 'Godzina wystąpienia porywu',
                'Minuta wystąpienia porywu', 'Temperatura gruntu -5 [°C]', 'Status pomiaru TG05',
                'Temperatura gruntu -10 [°C]', 'Status pomiaru TG10', 'Temperatura gruntu -20 [°C]',
                'Status pomiaru TG20', 'Temperatura gruntu -50 [°C]', 'Status pomiaru TG50',
                'Temperatura gruntu -100 [°C]',
                'Status pomiaru TG100', 'Temperatura minimalna za 12 godzin [°C]', 'Status pomiaru TMIN',
                'Temperatura maksymalna za 12 godzin [°C]', 'Status pomiaru TMAX',
                'Temperatura minimalna przy gruncie za 12 godzin [°C]', 'Status pomiaru TGMI',
                'Równoważnik wodny śniegu [mm/cm]', 'Status pomiaru RWSN', 'Wysokość pokrywy śnieżnej [cm]',
                'Status pomiaru PKSN', 'Wysokość świeżo spadłego śniegu [cm]', 'Status pomiaru HSS',
                'Wysokość śniegu na poletku [cm]', 'Status pomiaru GRSN', 'Gatunek śniegu [kod]',
                'Ukształtowanie pokrywy [kod]', 'Wysokość próbki [cm]', 'Status pomiaru HPRO', 'Ciężar próbki [g]',
                'Status pomiaru CIPR']


def search_for_keywords_in_columns(
        keywords, file_format=None
):
    """
    Search for the given keywords in the column names and return a dictionary with
    the file formats in which the keywords were found.

    Keyword arguments:
        keywords -- keywords that will be looked for
        file_format -- IMGW database file format, the columns of which will be
    used to look for keywords. If 'file_format' is None, then every column from
    every file will be taken and the function will show you exactly where the
    keywords were found (default None)
    """

    if file_format is None:
        intervals = ['monthly', 'daily', 'prompt']
        stations_kinds = ['synop', 'climat', 'fall']
        if type(keywords) == str:
            keywords = [keywords]

        found_file_formats = {}
        for interval in intervals:
            for kind in stations_kinds:
                if interval == 'prompt' and kind == 'fall':
                    continue
                file_formats = get_file_formats(interval, kind, 'all')
                found_file_formats['interval=' + str(interval) + ', ' + 'stations_kind=' + str(kind)] = file_formats

        keywords_in_files = {}
        if type(keywords) == list:
            for interval_stkind, file_formats in found_file_formats.items():
                for file in file_formats:
                    columns_names = get_column_names(file)
                    for name in columns_names:
                        for keyword in keywords:
                            if keyword.upper() in name.upper():
                                try:
                                    keywords_in_files[str(interval_stkind) + ", " + "file_format=" + str(file)].append(
                                        name)
                                except KeyError:
                                    keywords_in_files[str(interval_stkind) + ", " + "file_format=" + str(file)] = [name]
        else:
            raise ValueError("Invalid input for 'keywords'. Use a list of strings or a single str.")
        return keywords_in_files

    else:
        if type(file_format) == str:
            keywords_in_columns = []
            columns_names = get_column_names(file_format)
            for name in columns_names:
                if type(keywords) == list:
                    for keyword in keywords:
                        if keyword.upper() in name.upper():
                            keywords_in_columns.append(name)
                else:
                    raise ValueError("Invalid input for 'keywords'. Use a list of strings or a single str.")
        else:
            raise ValueError("Invalid input for the 'file_format' argument. Use a single str.")
        return keywords_in_columns


def get_urls(
        interval, stations_kind, years_range
):
    """
    Return the urls to the IMGW database for the given 'interval', 'stations_kind'
    and 'years_range'.

    Keyword arguments:
        interval -- data interval from the IMGW database (monthly, daily, prompt)
        stations_kind -- stations' kind (synop, climat, fall)
        years_range -- years range (e.g. range(1966, 2021))
    """

    if interval == 'monthly':
        interval = 'miesieczne/'
    elif interval == 'daily':
        interval = 'dobowe/'
    elif interval == 'prompt':
        interval = 'terminowe/'
        if stations_kind == 'fall':
            raise NotADirectoryError("There's no '/dane_meteorologiczne/terminowe/opad/' directory.")
    else:
        raise ValueError("Invalid 'interval' input. Available inputs: 'monthly', 'daily', 'prompt'.")

    if stations_kind == 'synop':
        stations_kind = 'synop/'
    elif stations_kind == 'fall':
        stations_kind = 'opad/'
    elif stations_kind == 'climat':
        stations_kind = 'klimat/'
    else:
        raise ValueError("Invalid 'stations_kind' input. Available inputs: 'synop', 'fall', 'climat'.")

    years_endings = []
    for year in years_range:
        if year < 2001:
            if year > 1995:
                if years_endings.count('1996_2000/') == 0:
                    years_endings.append('1996_2000/')
            elif year > 1990:
                if years_endings.count('1991_1995/') == 0:
                    years_endings.append('1991_1995/')
            elif year > 1985:
                if years_endings.count('1986_1990/') == 0:
                    years_endings.append('1986_1990/')
            elif year > 1980:
                if years_endings.count('1981_1985/') == 0:
                    years_endings.append('1981_1985/')
            elif year > 1975:
                if years_endings.count('1976_1980/') == 0:
                    years_endings.append('1976_1980/')
            elif year > 1970:
                if years_endings.count('1971_1975/') == 0:
                    years_endings.append('1971_1975/')
            elif year > 1965:
                if years_endings.count('1966_1970/') == 0:
                    years_endings.append('1966_1970/')
            elif year > 1959:
                if years_endings.count('1960_1965/') == 0:
                    years_endings.append('1960_1965/')
            else:
                raise ValueError("No data for {}. Available years range: 1960-2021.".format(str(year)))
        else:
            years_endings.append(str(year) + '/')

    urls = []
    base_url = 'https://danepubliczne.imgw.pl/data/dane_pomiarowo_obserwacyjne/dane_meteorologiczne/'
    for ending in years_endings:
        url = base_url + interval + stations_kind + ending
        urls.append(url)
    return urls


def download_data(urls):
    """
    Download data from the IMGW database.

    Keyword arguments:
        urls -- urls for the data which is requested
    """

    import requests
    from bs4 import BeautifulSoup as bs
    import zipfile
    import io

    files_reading_dir_path = str(__file__).replace('imgw.py', 'files_reading_folder')

    for url in urls:
        if urls.index(url) == 0:
            print("Starting data download... 0% done")

        r = requests.get(url)
        soup = bs(r.content, features="html.parser")

        zip_file_paths = []
        for element in soup.find_all('a'):
            if '.zip' in element.get_text():
                zip_file_paths.append(element.get_text())

        for path in zip_file_paths:
            zip_file = requests.get(url + path)
            zip_file = zipfile.ZipFile(io.BytesIO(zip_file.content))
            zip_file.extractall(files_reading_dir_path)

        if (urls.index(url) + 1) % 5 == 0:
            print("Downloading data... {}% done".format(round((urls.index(url) / len(urls)) * 100)))
    print("Data downloaded! 100% done")


def concatenate_data(
        downloaded_files_names, file_formats, specific_columns,
        keywords, optimize_memory_usage, years_range,
        merge_splitted_stations
):
    """
    Merge tables from downloaded files and return them as one merged pd.DataFrame.

    Keyword arguments:
        downloaded_files_names -- list of downloaded file names
        file_formats -- IMGW file formats included in downloaded files
        specific_columns -- which columns will be taken for merge
        keywords -- words which must be in the column name if the column is to be
    merged
        optimize_memory_usage -- reduce pd.DataFrame memory usage
        years_range -- filter pd.DataFrame up to the given period
    """

    import pandas as pd
    import numpy as np
    import os

    if isinstance(file_formats, list) and len(file_formats) > 1:
        raise ValueError(
            f"""Invalid value for the 'file_format' argument. Data downloading is possible only for a single file. 
            {len(file_formats)} files given ({file_formats}).
            """
        )

    if isinstance(file_formats, str):
        file_formats = [file_formats]

    df = pd.DataFrame()
    keywords_in_columns = []
    files_reading_dir_path = str(__file__).replace('imgw.py', 'files_reading_folder')

    for file in downloaded_files_names:
        if downloaded_files_names.index(file) == 0:
            print("Starting data concatenating... 0% done")

        for file_format in file_formats:

            if file_format == 's_d' or file_format == 'k_d':
                avoid_file_format = file_format + '_t'
            else:
                avoid_file_format = 'NO FILE FORMAT TO AVOID'

            if file_format in file and avoid_file_format not in file:

                path = os.path.join(files_reading_dir_path, file)
                csv_DataFrame = pd.read_csv(path, encoding="windows-1250", header=None)

                if specific_columns is not None and keywords is None:
                    if type(specific_columns) == str:
                        specific_columns = [specific_columns]

                    if type(specific_columns) is list:
                        csv_DataFrame = csv_DataFrame[specific_columns]
                    else:
                        raise ValueError("Invalid 'specific_columns' type. Use a list of strs or a single str.")

                if keywords is not None and len(keywords_in_columns) == 0:
                    columns = get_column_names(file_format)
                    for keyword in keywords:
                        for column in columns:
                            if keyword.upper() in column.upper():
                                keywords_in_columns.append(columns.index(column))

                if keywords is not None:
                    csv_DataFrame = csv_DataFrame[keywords_in_columns]

                if optimize_memory_usage:
                    for column in csv_DataFrame.columns:
                        if column == 0:
                            csv_DataFrame[column] = csv_DataFrame[column].astype('object', errors='ignore')
                        elif csv_DataFrame[column].dtype == np.dtype('int64'):
                            csv_DataFrame[column] = csv_DataFrame[column].astype('int16', errors='ignore')
                        elif csv_DataFrame[column].dtype == np.dtype('float64'):
                            csv_DataFrame[column] = csv_DataFrame[column].astype('float16', errors='ignore')
                        else:
                            csv_DataFrame[column] = csv_DataFrame[column].astype('float16', errors='ignore')

                df = df.append(csv_DataFrame)

        try:
            if (downloaded_files_names.index(file) + 1) % (round(len(downloaded_files_names) * 0.1)) == 0:
                print("Concatenating data... {}% done".format(
                    round((downloaded_files_names.index(file) / len(downloaded_files_names)) * 100)))
        except ZeroDivisionError:
            pass
    print("Data concatenated! 100% done \n")

    try:
        df = df[df[2] >= min(
            years_range)]
        df = df[df[2] <= max(
            years_range)]
    except KeyError:
        print("No column with year has been specified. Can't limit data range to given 'years_range'.")

    if merge_splitted_stations and 1 in df.columns:
        stations_series = df[1]
        new_stations_series = []
        for station in stations_series:
            if station == 'KATOWICE-MUCHOWIEC':
                new_stations_series.append('KATOWICE')
            elif station == 'ŁÓDŹ-LUBLINEK':
                new_stations_series.append('ŁÓDŹ')
            elif station == 'POZNAŃ-ŁAWICA':
                new_stations_series.append('POZNAŃ')
            elif station == 'WARSZAWA-OKĘCIE':
                new_stations_series.append('WARSZAWA')
            elif station == 'WROCŁAW-STRACHOWICE':
                new_stations_series.append('WROCŁAW')
            elif station == 'ELBLĄG-MILEJEWO':
                new_stations_series.append('ELBLĄG')
            elif station == 'RESKO-SMÓLSKO':
                new_stations_series.append('RESKO')
            elif station == 'KOŁOBRZEG-DŹWIRZYNO':
                new_stations_series.append('KOŁOBRZEG')
            else:
                new_stations_series.append(station)
        df[1] = new_stations_series

    if keywords is not None:
        return df, keywords_in_columns
    else:
        return df


def download_imgw_climatological_data(
        interval, stations_kind, years_range,
        file_format_index=0, file_format=None, specific_columns=None,
        keywords=None, merge_splitted_stations=True, optimize_memory_usage=False,
        return_coordinates=False
):
    """
    Download the IMGW data files and return them as one merged pd.DataFrame.

    Keyword arguments:
        interval -- data interval from the IMGW database ('monthly', 'daily',
    'prompt')
        stations_kind -- stations' kind ('synop', 'climat', 'fall')
        years_range -- years range (e.g. range(2010, 2021))
        file_format_index -- for which element (from the list of the available file
    formats in the IMGW database) the data will be downloaded. Usually the length of
    the list is 2. Valid inputs: 0, 1, 'all' (default 0)
        file_format -- IMGW database file format. The argument has priority over
    'file_format_index'. If None, the data will be downloaded correspondingly to
    'file_format_index' (default None)
        specific_columns -- which columns from downloaded files will be taken
    to merge. If None, take all columns (default None)
        keywords -- words which must be in the column name if the column is to be
    merged. If None, do not filter the column names (default None)
        merge_splitted_stations -- merge stations which are the same but have
    different names (default True)
        optimize_memory_usage -- reduce pd.DataFrame memory usage (default False)
        return_coordinates -- add columns with latitude, longitude and elevation
    to the returned DataFrame (default False)
    """

    from os import listdir
    from os.path import isfile, join
    import shutil

    if file_format not in get_file_formats(interval, stations_kind, 'all') and file_format is not None:
        raise ValueError(
            f"""
            There's no such file format ({file_format}) for the specified combination of 'interval' and 'stations_kind'
            """
        )

    files_reading_dir_path = str(__file__).replace('imgw.py', 'files_reading_folder')

    urls = get_urls(interval, stations_kind, years_range)
    download_data(urls)
    downloaded_files_names = [f for f in listdir(files_reading_dir_path) if isfile(join(files_reading_dir_path, f))]

    if file_format is not None:
        if type(file_format) == str:
            file_formats = [file_format]
        elif type(file_format) == list:
            file_formats = [format_ for format_ in file_format]
        else:
            raise ValueError("Invalid input for the 'file_format' argument. Use a signle str or a list of strs.")
    else:
        file_formats = get_file_formats(interval, stations_kind, file_format_index)

    df = concatenate_data(downloaded_files_names, file_formats, specific_columns,
                          keywords, optimize_memory_usage, years_range,
                          merge_splitted_stations)

    if keywords is not None:
        keywords_in_columns = df[1]
        df = df[0]
    else:
        df = df
        keywords_in_columns = None

    if len(file_formats) == 1:
        column_names = get_column_names(file_formats[0])

        if specific_columns is not None and keywords is None:
            chosen_columns_names = []
            if type(specific_columns) == list:
                for column_index in specific_columns:
                    chosen_columns_names.append(column_names[column_index])
            df.columns = chosen_columns_names

        elif keywords is not None:
            chosen_columns_names = []
            for column_index in keywords_in_columns:
                chosen_columns_names.append(column_names[column_index])
            df.columns = chosen_columns_names

        else:
            df.columns = column_names

    shutil.rmtree(files_reading_dir_path)

    if return_coordinates:

        import pandas as pd

        lat = []
        lon = []
        elv = []

        path = str(__file__).replace('imgw.py', 'imgw_coordinates.csv')
        station_coordinates = pd.read_csv(path, index_col=0)

        for station in list(df['Nazwa stacji']):
            try:
                coordinates = station_coordinates[station]
            except KeyError:
                lat.append(None)
                lon.append(None)
                elv.append(None)
                continue

            lat.append(coordinates[0])
            lon.append(coordinates[1])
            elv.append(coordinates[2])

        df['lat'] = lat
        df['lon'] = lon
        df['elv'] = elv

    return df



import cloupy as cl
import pytest
import matplotlib.pyplot as plt
import urllib.request
import urllib.error


def check_if_NOT_connected_to_the_internet(host='http://google.com'):
    try:
        urllib.request.urlopen(host)
        return False
    except urllib.error.URLError:
        return True


@pytest.mark.skipif(check_if_NOT_connected_to_the_internet(), reason='internet connection required')
class TestDrawingWithDailyData:

    def test_drawing_and_daily_data_downloading(self):
        plotted_figures_before = plt.gcf().number
        wl = cl.g_WalterLieth('POZNAŃ')
        wl.d_imgw_data(interval='daily', stations_kind='synop', years_range=range(2010, 2011))
        wl.draw()
        plotted_figures_after = plt.gcf().number

        assert plotted_figures_before < plotted_figures_after

    def test_daily_data_downloading_and_drawing_from_global_df(self):
        data = cl.d_imgw_data('daily', 'synop', range(2010, 2011))
        cl.set_global_df(data)

        plotted_figures_before = plt.gcf().number
        assert TestDrawingWithDailyData.draw_from_global_df('WARSZAWA') > plotted_figures_before

        plotted_figures_before = plt.gcf().number
        assert TestDrawingWithDailyData.draw_from_global_df('KOŁOBRZEG') > plotted_figures_before

        plotted_figures_before = plt.gcf().number
        assert TestDrawingWithDailyData.draw_from_global_df('KOŁO') > plotted_figures_before

    @staticmethod
    def draw_from_global_df(station_name):
        wl = cl.g_WalterLieth(station_name)
        wl.import_global_df('imgw_daily')
        wl.draw()
        return plt.gcf().number



from cloupy.scraping import imgw
import pytest


class TestFileFormats:
    def test_monthly_arg(self):
        # stations_kind = 'synop'
        assert imgw.get_file_formats('monthly', 'synop', 0) == ['s_m_d']
        assert imgw.get_file_formats('monthly', 'synop', 1) == ['s_m_t']
        assert imgw.get_file_formats('monthly', 'synop', 'all') == ['s_m_d', 's_m_t']

        # stations_kind = 'fall'
        for file_format_index in [0, 1, 'all']:
            assert imgw.get_file_formats('monthly', 'fall', file_format_index) == ['o_m']

        # stations_kind = 'climat'
        assert imgw.get_file_formats('monthly', 'climat', 0) == ['k_m_d']
        assert imgw.get_file_formats('monthly', 'climat', 1) == ['k_m_t']
        assert imgw.get_file_formats('monthly', 'climat', 'all') == ['k_m_d', 'k_m_t']

        # stations_kind is invalid
        with pytest.raises(ValueError):
            imgw.get_file_formats('monthly', 'abc', 0)

    def test_daily_arg(self):
        # stations_kind = 'synop'
        assert imgw.get_file_formats('daily', 'synop', 0) == ['s_d']
        assert imgw.get_file_formats('daily', 'synop', 1) == ['s_d_t']
        assert imgw.get_file_formats('daily', 'synop', 'all') == ['s_d', 's_d_t']

        # stations_kind = 'fall'
        for file_format_index in [0, 1, 'all']:
            assert imgw.get_file_formats('daily', 'fall', file_format_index) == ['o_d']

        # stations_kind = 'climat'
        assert imgw.get_file_formats('daily', 'climat', 0) == ['k_d']
        assert imgw.get_file_formats('daily', 'climat', 1) == ['k_d_t']
        assert imgw.get_file_formats('daily', 'climat', 'all') == ['k_d', 'k_d_t']

        # stations_kind is invalid
        with pytest.raises(ValueError):
            imgw.get_file_formats('daily', 'abc', 0)

    def test_prompt_arg(self):
        # stations_kind = 'synop' and 'climat'
        for file_format_index in [0, 1, 'all']:
            assert imgw.get_file_formats('prompt', 'synop', file_format_index) == ['s_t']
            assert imgw.get_file_formats('prompt', 'climat', file_format_index) == ['k_t']

        # stations_kind = 'fall'
        with pytest.raises(NotADirectoryError):
            imgw.get_file_formats('prompt', 'fall', 0)

        # stations_kind is invalid
        with pytest.raises(ValueError):
            imgw.get_file_formats('prompt', 'abc', 0)


class TestKeywordsInColumns:
    def test_None_file_format(self):
        lowercased_keywords = ['temperatura', 'opad']
        uppercased_keywords = ['TEMPERATURA', 'OPAD']
        capitalized_keywords = ['Temperatura', 'Opad']

        lowercased_result = TestKeywordsInColumns.values_counter(
            imgw.search_for_keywords_in_columns(lowercased_keywords)
        )

        uppercased_result = TestKeywordsInColumns.values_counter(
            imgw.search_for_keywords_in_columns(uppercased_keywords)
        )

        capitalized_result = TestKeywordsInColumns.values_counter(
            imgw.search_for_keywords_in_columns(capitalized_keywords)
        )

        assert lowercased_result == uppercased_result == capitalized_result

    def test_specific_file_format(self):
        lowercased_keywords = ['temperatura', 'opad']
        uppercased_keywords = ['TEMPERATURA', 'OPAD']
        capitalized_keywords = ['Temperatura', 'Opad']

        expected_lengths = {
            'k_m_d': 12,
            'k_m_t': 1,
            'o_m': 5,
            's_m_d': 13,
            's_m_t': 3,
            'k_d': 6,
            'k_d_t': 1,
            'o_d': 2,
            's_d': 9,
            's_d_t': 3,
            'k_t': 2,
            's_t': 13
        }
        for format_, expected_value in expected_lengths.items():
            lowercased_result = len(imgw.search_for_keywords_in_columns(lowercased_keywords, format_))

            uppercased_result = len(imgw.search_for_keywords_in_columns(uppercased_keywords, format_))

            capitalized_result = len(imgw.search_for_keywords_in_columns(capitalized_keywords, format_))

            assert lowercased_result == uppercased_result == capitalized_result == expected_lengths[format_]

    @staticmethod
    def values_counter(dictionary):
        counter = 0
        for key in dictionary:
            counter += len(dictionary[key])
        return counter



from cloupy.diagrams.walter_lieth import WalterLieth
import pytest
import pandas as pd
import matplotlib.pyplot as plt


class TestDrawing:

    @pytest.fixture
    def data_for_humid_period(self):
        return pd.DataFrame(
            {
                'months': {
                    0: 1, 1: 2, 2: 3, 3: 4, 4: 5, 5: 6, 6: 7, 7: 8, 8: 9, 9: 10,
                    10: 11, 11: 12
                },
                'temp': {
                    0: -1.36, 1: -0.59, 2: 3.03, 3: 8.35, 4: 13.55, 5: 17.02, 6: 18.64,
                    7: 18.07, 8: 13.76, 9: 8.82, 10: 3.87, 11: 0.36
                },
                'preci': {
                    0: 32.70, 1: 26.76, 2: 31.40, 3: 33.63, 4: 51.10, 5: 61.09,
                    6: 79.37, 7: 57.99, 8: 42.59, 9: 37.12, 10: 37.18, 11: 38.74
                },
                'temp_max': {
                    0: 5.45, 1: 6.79, 2: 13.54, 3: 20.07, 4: 24.67, 5: 28.25, 6: 29.65,
                    7: 29.21, 8: 25.24, 9: 18.88, 10: 11.63, 11: 7.66
                },
                'temp_min': {
                    0: -11.22, 1: -10.60, 2: -6.38, 3: -1.66, 4: 2.95, 5: 6.77, 6: 9.38,
                    7: 8.72, 8: 4.23, 9: 0.31, 10: -3.72, 11: -8.56
                }
            }
        )

    @pytest.fixture
    def default_settings(self):
        return {
            'figsize': (7.74, 7.74), 'language': None, 'freeze_rectangles': True,
            'title_text': True, 'years_text': True, 'coordinates_box': True,
            'yearly_means_box': True, 'extremes_box': True, 'legend_box': True
        }

    @pytest.fixture
    def opposite_settings(self):
        return {
            'figsize': (7.74, 7.74), 'language': 'POL', 'freeze_rectangles': False,
            'title_text': False, 'years_text': False, 'coordinates_box': False,
            'yearly_means_box': False, 'extremes_box': False, 'legend_box': False
        }

    def test_humid_period_and_elements_viewing(
            self, default_settings, opposite_settings,
            data_for_humid_period
    ):
        for setting, value in opposite_settings.items():
            default_settings[setting] = value

            plotted_figures_before = plt.gcf().number
            wl = WalterLieth(
                'POZNAŃ', dataframe=data_for_humid_period, years_range=range(1951, 2020),
                lat=52.42, lon=16.83, elevation=92
            )
            wl.draw(
                figsize=default_settings['figsize'], language=default_settings['language'],
                freeze_rectangles=default_settings['freeze_rectangles'], title_text=default_settings['title_text'],
                years_text=default_settings['years_text'], coordinates_box=default_settings['coordinates_box'],
                yearly_means_box=default_settings['yearly_means_box'], extremes_box=default_settings['extremes_box'],
                legend_box=default_settings['legend_box']
            )
            plotted_figures_after = plt.gcf().number

            assert plotted_figures_before < plotted_figures_after

    def test_dry_period(
            self, data_for_humid_period
    ):
        data_for_dry_period = data_for_humid_period
        data_for_dry_period.preci = [0] * 12

        plotted_figures_before = plt.gcf().number
        wl = WalterLieth(
            'POZNAŃ', dataframe=data_for_dry_period, years_range=range(1951, 2020),
            lat=52.42, lon=16.83, elevation=92
        )
        wl.draw()
        plotted_figures_after = plt.gcf().number

        assert plotted_figures_before < plotted_figures_after

    def test_mixed_all_periods(
            self, data_for_humid_period
    ):
        data_for_mixed_periods = data_for_humid_period
        data_for_mixed_periods['preci'] = [110, 70, 50, 40, 30, 20, 0, 0, 20, 30, 70, 110]

        plotted_figures_before = plt.gcf().number
        wl = WalterLieth(
            'POZNAŃ', dataframe=data_for_mixed_periods, years_range=range(1951, 2020),
            lat=52.42, lon=16.83, elevation=92
        )
        wl.draw()
        plotted_figures_after = plt.gcf().number

        assert plotted_figures_before < plotted_figures_after



"""
Enjoy easy data scraping from websites and its processing. What's more, enjoy
even easier data visualization!

-----DATA PROCESSING FUNCTIONS/CLASSES----
    set_global_df() -- set a global dataframe from which data can be imported at
any time and place
    read_global_df() -- return the global data frame as pandas.DataFrame

    DataFrame() - create pandas.DataFrame object
-------------------------------------------

---------DATA SCRAPING FUNCTIONS--------
    d_imgw_data() -- download IMGW data files from the IMGW database and return them
as one merged pd.DataFrame
    i_imgw_get_file_formats() -- return the available file formats for the given
'interval' and 'stations_kind' in the IMGW database (different file formats contain
different data)
    i_imgw_search_keywords_in_columns() -- search for the given keywords in the column
names and return a dictionary with the file formats in which the keywords were found

    d_wmo_data() -- download climatological data for specified station/stations
from the WMO website
    i_wmo_get_stations() -- return pandas.DataFrame with WMO stations information
(WMO ids, coordinates, etc.)
    i_wmo_search_near_station() -- return the nearest stations from the WMO database
for the given coordinates
----------------------------------------

-------DATA VISUALIZATION FUNCTIONS------
    choose_diagStyle() -- choose a global style for diagrams
    change_diagStyle_params() -- change global parameters for drawing diagrams

    save_graph() -- save created graph (it is the function 'savefig' from the
    matplotlib library)
-----------------------------------------

--------DATA VISUALIZATION CLASSES-------
# Note that every class for drawing diagrams contains some of the above functions
as its methods (for data scraping and processing)

    g_WalterLieth() -- create a WalterLieth object in which data for drawing a
Walter-Lieth diagram can be downloaded, modified, manually provided
-----------------------------------------
"""


import pandas as pd
import os
from cycler import cycler
from matplotlib.pyplot import *
from matplotlib import rcParams

from cloupy.scraping.imgw import download_imgw_climatological_data as d_imgw_data
from cloupy.scraping.imgw import get_file_formats as i_imgw_get_file_formats
from cloupy.scraping.imgw import search_for_keywords_in_columns as i_imgw_search_keywords_in_columns

from cloupy.scraping.wmo import download_wmo_climatological_data as d_wmo_data
from cloupy.scraping.wmo import get_wmo_stations_info as i_wmo_get_stations
from cloupy.scraping.wmo import search_for_the_nearest_station as i_wmo_search_near_station

from cloupy.diagrams.walter_lieth import WalterLieth as g_WalterLieth

from pandas import DataFrame as DataFrame
from matplotlib.pyplot import savefig as save_graph

try:  # delete global data frame from the previous session
    path_ = str(__file__).replace('__init__.py', '')

    os.remove(path_ + r'global_df.csv')
    os.remove(path_ + r'global_df.json')
except FileNotFoundError:
    pass

#  dictionaries of styles for drawing
default_style = {
    'font.family': 'Calibri',
    'axes.titleweight': 'bold',
    'axes.labelweight': 'bold',
    'figure.figsize': (6, 6 / 1.618),
    'axes.edgecolor': 'black',
    'axes.labelsize': 10,
    'axes.prop_cycle':
        (
                cycler(color=['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
                              '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']) +
                cycler(linestyle=['-' for solid in range(0, 10)])
        )
}

retro_style = {
    'font.family': 'Times New Roman',
    'axes.titleweight': 'bold',
    'axes.labelweight': 'light',
    'figure.figsize': (6, 6 / 1.618),
    'axes.edgecolor': 'black',
    'axes.labelsize': 12,
    'axes.prop_cycle':
        (
                cycler(color=['k' for k in range(0, 6)]) +
                cycler(linestyle=['-', '--', '-.',
                                  (0, (1, 5)),
                                  (0, (3, 1, 1, 1)),
                                  (0, (3, 10, 1, 10, 1, 10))
                                  ]
                       )
        )
}

#  set default style for cloudy
diagStyle = rcParams
diagStyle['savefig.dpi'] = 300
diagStyle['savefig.bbox'] = 'tight'
diagStyle['savefig.pad_inches'] = 0.3
for param, value in default_style.items():
    diagStyle[param] = value


def change_diagStyle_params(diagStyle_dict):
    """
    Change global parameters for drawing diagrams.

    Keyword arguments:
        diagStyle_dict -- a dictionary in which the keys are parameters that have
    to be changed; values are the values to which default values have to be changed

    ---------------NOTE THAT---------------
    For available parameters and their valid values you can check matplotlib's
    'rcParams' documentation:
    https://matplotlib.org/stable/tutorials/introductory/customizing.html
    ---------------------------------------
    """

    global diagStyle
    for changed_param, changed_value in diagStyle_dict.items():
        diagStyle[changed_param] = changed_value


def choose_diagStyle(diag_style='default'):
    """
    Choose a global style for diagrams.

    Keyword arguments:
        diag_style -- a style for diagrams. Available styles: 'default', 'retro'
    (default 'default')
    """
    if diag_style == 'default':
        for default_param, default_value in default_style.items():
            diagStyle[default_param] = default_value
    elif diag_style == 'retro':
        for retro_param, retro_value in retro_style.items():
            diagStyle[retro_param] = retro_value
    else:
        raise ValueError(
            "Invalid 'diag_style' argument. Available arguments: 'default', 'retro'"
        )


def set_global_df(
        pd_DataFrame
):
    """
    Set global dataframe from which data can be imported at any time and place.

    Keyword arguments:
        pd_DataFrame -- a pandas DataFrame object which will be the global DataFrame
    """

    path = str(__file__).replace('__init__.py', '')

    try:
        os.remove(path + r'global_df.csv')
    except FileNotFoundError:
        pass

    if not isinstance(pd_DataFrame, pd.DataFrame):
        raise ValueError(
            """
            Invalid object type for 'pd_DataFrame' which must be 'pandas.DataFrame'. More info on creating pandas.DataFrame() 
            object is available here: 'https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html'
            """
        )

    pd_DataFrame.to_csv(path + r'global_df.csv')


def read_global_df():
    """
    Return the global dataframe which was previously set by the 'set_global_df'
    function.
    """
    path = str(__file__).replace('__init__.py', '')

    try:
        df = pd.read_csv(path + r'global_df.csv', index_col=0)
    except FileNotFoundError:
        raise FileNotFoundError(
            "Use cloudy.set_global_df to set a global dataframe and then use cloudy.read_global_df again.")
    else:
        return df



from cloupy.scraping import imgw
import pytest
import urllib.request
import urllib.error


def check_if_NOT_connected_to_the_internet(host='http://google.com'):
    try:
        urllib.request.urlopen(host)
        return False
    except urllib.error.URLError:
        return True


@pytest.mark.filterwarnings("ignore::pandas.errors.DtypeWarning")
@pytest.mark.skipif(check_if_NOT_connected_to_the_internet(), reason='internet connection required')
class TestDataDownloading:

    @pytest.fixture
    def intervals(self):
        return ['monthly', 'daily', 'prompt']

    @pytest.fixture
    def st_kinds(self):
        return ['synop', 'climat', 'fall']

    def test_if_column_2_is_always_year(
            self, intervals, st_kinds
    ):

        from os import listdir
        from os.path import isfile, join
        import shutil
        from random import shuffle

        y_range = range(2018, 2019)
        files_reading_dir_path = str(__file__).replace(
            join('test', 'test_integration', 'test_integration_imgw.py'),
            join('scraping', 'files_reading_folder')
        )

        for interval in intervals:
            for st_kind in st_kinds:

                if st_kind == 'fall' and interval == 'prompt':
                    continue

                urls = imgw.get_urls(interval, st_kind, y_range)
                imgw.download_data(urls)
                downloaded_files_names = [f for f in listdir(files_reading_dir_path) if
                                          isfile(join(files_reading_dir_path, f))]

                file_formats = imgw.get_file_formats(interval, st_kind, 'all')
                keywords = ['nazwa stacji', 'temperatura', 'rok', 'opad', 'wiatr']
                shuffle(keywords)

                for file in file_formats:

                    if isinstance(file_formats, str):
                        file = file_formats

                    df = imgw.concatenate_data(
                        downloaded_files_names=downloaded_files_names, file_formats=file, years_range=y_range,
                        keywords=keywords, specific_columns=None, optimize_memory_usage=False,
                        merge_splitted_stations=True
                    )

                    df = df[0][df[1]]

                    assert min(df[2]) == 2018

                shutil.rmtree(files_reading_dir_path)

    def test_data_downloading_for_years_before_2001(
            self, intervals, st_kinds
    ):
        years_range = range(1984, 1987)
        TestDataDownloading.download_and_test_data(intervals, st_kinds, years_range)

    def test_data_downloading_for_years_after_2000(
            self, intervals, st_kinds
    ):
        years_range = range(2011, 2013)
        TestDataDownloading.download_and_test_data(intervals, st_kinds, years_range)

    def test_data_downloading_for_years_between_2000_and_2001(
            self, intervals, st_kinds
    ):
        years_range = range(2000, 2002)
        TestDataDownloading.download_and_test_data(intervals, st_kinds, years_range)

    def test_adding_coordinates_to_dataframe(
            self, intervals, st_kinds
    ):
        years_range = range(2010, 2011)
        for interval in intervals:
            for st_kind in st_kinds:

                if st_kind == 'fall' and interval == 'prompt':
                    continue

                df = imgw.download_imgw_climatological_data(
                    interval, st_kind, years_range,
                    specific_columns=[0, 1, 2, 3],
                    optimize_memory_usage=True,
                    return_coordinates=True
                )

                assert 'lat' in df.columns
                assert 'lon' in df.columns
                assert 'elv' in df.columns

                assert not df['lat'].isnull().all()
                assert not df['lon'].isnull().all()
                assert not df['elv'].isnull().all()

    @staticmethod
    def download_and_test_data(
            intervals, st_kinds, years_range
    ):
        for interval in intervals:
            for st_kind in st_kinds:
                if interval == 'prompt' and st_kind == 'fall':
                    with pytest.raises(NotADirectoryError):
                        imgw.download_imgw_climatological_data(
                            interval, st_kind, years_range
                        )
                        continue
                else:
                    df = imgw.download_imgw_climatological_data(
                        interval, st_kind, years_range,
                        optimize_memory_usage=True,
                        specific_columns=[0, 1, 2, 3]
                    )

                assert not df.empty



from setuptools import setup, find_packages

VERSION = '1.0.0'
DESCRIPTION = 'The package allows to download, process and visualize climatological data from reliable sources'
README = open('README.md', 'r', encoding='utf8').read()

setup(
    name='cloupy',
    version=VERSION,
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type='text/markdown',
    author='Kamil Grala',
    author_email='kamil.grala32466@gmail.com',
    url='https://github.com/pdGruby/cloupy',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'pandas>=1.3.3,<=1.3.4',
        'matplotlib>=3.4.3,<=3.4.3',
        'requests>=2.26.0,<=2.26.0',
        'beautifulsoup4>=4.9.3,<=4.9.3',
        'numpy>=1.21.4,<=1.21.4'
    ],
    tests_require=[
        'pytest>=6.2.5',
        'mock>=4.0.3'
    ],
    package_data={
        'cloupy': [
            'scraping/*',
            'diagrams/*',
            'test/test_integration/*',
            'test/test_unit/*'
        ],
                  },
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: POSIX :: Linux',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Framework :: Matplotlib',
        'Topic :: Scientific/Engineering :: Atmospheric Science'
    ]
)